#include <Arduino.h>


#include <TimeLib.h>                              //By Michael Margolis
#include <Servo.h>
#include <Wire.h>
#include <DueFlashStorage.h>
#include <DS1307RTC.h>                            //By Machael Margolis

/*
   Copyright Sergio Daniel Castaneda Nino 2018
   PLC Arduino Firmware DUE Bluetooth V1.0

   IMPORTANT: Everytime you load this firmware to your Arduino DUE the previous ladder design 
              that's in the board will be deleted so you should load the ladder design whit the PLC Ladder Simulator app again.
*/


#define BYTE6MASK0 0x3F
#define BYTE6MASK1 0xFC
#define BYTE6MASK2 0X3F000
#define BYTE6MASK3 0xFC0000
#define BYTE6MASK4 0x3F000000
#define BYTE6MASK5 0xC0000000

#define GMASK8 0xFF;
#define GMASK16 0xFFFF;

#define MDMASK 0XFC0
#define HDMASK 0xF000

#define BYTEMASK0 0xFF
#define BYTEMASK1 0xFF00
#define BYTEMASK2 0xFF0000
#define BYTEMASK3 0xFF000000

#define REG_INT 23101
#define REG_LONG 23230
#define REG_FLOAT 23487
#define REG_TIMER 23744
#define REG_COUNTER 24033
#define REG_STATUS 24418
#define REG_IN 24419
#define REG_OUT 24435
#define REG_VIRTUAL 24453
#define REG_TYPE_SIZE 24517


DueFlashStorage dueFlashStorage;
                                                         

/*
 * HMI MODE
 * HMI mode allows you to connect your board using this firmware with the HMI Controller for Arduino aplication for Android.
 * 
 * When the HMI mode is enabled the user only has 15 seconds to load a new ladder desing to the board once the board is 
 * turn it on. After the 15 seconds the board can connect with the HMI app.
 * 
 */
bool enableHMImode = false;                                                            //SET TO TRUE IF USING HMI CONTROLLER FOR ARDUINO APP


bool enabledRTC = false;                                                               //SET TO TRUE IF USING A RTC MODULE (REAL-TIME-CLOCK)
int clockGMT = -500;

/* SET TIME ZONE GMT (GREENWICH MEAN TIME)
 * Set your time zone GMT in the "clockGMT" variable. 
 * Example 1: if your time zone is GMT+2:00 set clockGMT = 200;
 * Example 2: if your time zone is GMT-5:00 set clockGMT = -500;
 * Example 3: if your time zonie is GMT+9:30 set clockGMT = 930;
 * 
 * Name   Description                         Relative to GMT
 * GMT    Greenwich Mean Time                       GMT
 * UTC    Universal Coordinated Time                GMT
 * ECT    European Central Time                     GMT+1:00
 * EET    Eastern European Time                     GMT+2:00
 * ART    (Arabic) Egypt Standard Time              GMT+2:00
 * EAT    Eastern African Time                      GMT+3:00
 * MET    Middle East Time                          GMT+3:30
 * NET    Near East Time                            GMT+4:00
 * PLT    Pakistan Lahore Time                      GMT+5:00
 * IST    India Standard Time                       GMT+5:30
 * BST    Bangladesh Standard Time                  GMT+6:00
 * VST    Vietnam Standard Time                     GMT+7:00
 * CTT    China Taiwan Time                         GMT+8:00
 * JST    Japan Standard Time                       GMT+9:00
 * ACT    Australia Central Time                    GMT+9:30
 * AET    Australia Eastern Time                    GMT+10:00
 * SST    Solomon Standard Time                     GMT+11:00
 * NST    New Zealand Standard Time                 GMT+12:00
 * MIT    Midway Islands Time                       GMT-11:00
 * HST    Hawaii Standard Time                      GMT-10:00
 * AST    Alaska Standard Time                      GMT-9:00
 * PST    Pacific Standard Time                     GMT-8:00
 * PNT    Phoenix Standard Time                     GMT-7:00
 * MST    Mountain Standard Time                    GMT-7:00
 * CST    Central Standard Time                     GMT-6:00
 * EST    Eastern Standard Time                     GMT-5:00
 * IET    Indiana Eastern Standard Time             GMT-5:00
 * PRT    Puerto Rico and US Virgin Islands Time    GMT-4:00
 * CNT    Canada Newfoundland Time                  GMT-3:30
 * AGT    Argentina Standard Time                   GMT-3:00
 * BET    Brazil Eastern Time                       GMT-3:00
 * CAT    Central African Time                      GMT-1:00
 */


unsigned long currentGetRTCUpdateTime;
const int targetGetRTCUpdateTime = 300;


unsigned long startReprogramTime;
unsigned long startBlinkReproTime;
const unsigned long targetReprogramTime = 15000;
const unsigned long blinkTime = 200;

unsigned long startTimeSend;
unsigned long targetTimeSend = 1000;

bool inReprogramTime = true;
bool runReprogramTime = false;

bool sendStartSend = false;

bool alreadySendIt = false;
bool inProcessLock = true;
bool requestSecondPart = false;
bool inSecondState = false;

bool iState[16]= {false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false};
bool oState[18]= {false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false};
bool vState[64]= {false, false, false, false, false, false, false, false,false, false, false, false, false, false, false, false,false, false, false, false, false, false, false, false,false, false, false, false, false, false, false, false,
                  false, false, false, false, false, false, false, false,false, false, false, false, false, false, false, false,false, false, false, false, false, false, false, false,false, false, false, false, false, false, false, false};

bool iAnalog[12] = {false, false, false, false, false, false, false, false, false, false, false, false};
bool oAnalog[8] = {false, false, false, false, false, false, false, false};

bool iPreState[16] = {false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false};

bool oPreState[18] = {false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false};
bool vPreState[64] = {false, false, false, false, false, false, false, false,false, false, false, false, false, false, false, false,false, false, false, false, false, false, false, false,false, false, false, false, false, false, false, false,
                      false, false, false, false, false, false, false, false,false, false, false, false, false, false, false, false,false, false, false, false, false, false, false, false,false, false, false, false, false, false, false, false,};

bool dacAnalogMask[2] = {false, false};

bool iMask[16]= {false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false}; //0-7
bool oMask[18]= {false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false};
bool vMask[64]= {false, false, false, false, false, false, false, false,false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false,false, false, false, false, false, false, false, false,
                false, false, false, false, false, false, false, false,false, false, false, false, false, false, false, false,false, false, false, false, false, false, false, false,false, false, false, false, false, false, false, false};

bool iUpEdge[16]= {false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false}; //0-7
bool oUpEdge[18]= {false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false};
bool vUpEdge[64]= {false, false, false, false, false, false, false, false,false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false,false, false, false, false, false, false, false, false,
                                false, false, false, false, false, false, false, false,false, false, false, false, false, false, false, false,false, false, false, false, false, false, false, false,false, false, false, false, false, false, false, false};

bool iDownEdge[16]= {false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false}; //0-7
bool oDownEdge[18]= {false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false};
bool vDownEdge[64]= {false, false, false, false, false, false, false, false,false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false,false, false, false, false, false, false, false, false,
                                false, false, false, false, false, false, false, false,false, false, false, false, false, false, false, false,false, false, false, false, false, false, false, false,false, false, false, false, false, false, false, false};

int16_t intValues[64] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
int32_t dintValues[64]= {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
float realValues[64]  = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

int8_t iIndex[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
int8_t oIndex[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
int8_t vIndex[64] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
int8_t intIndex[64] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
int8_t dintIndex[64] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
int8_t realIndex[64] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

uint8_t iPins[16] = {A0, A1, A2, A3, A4, A5, A6, A7, A8, A9, A10, A11, 48, 49, 50, 51}; //INPUT pins
uint8_t oPins[18] = {3, 5, 6, 7, 8, 9, 11, 12, 23, 24, 25, 26, 27, 28, 29, 30, DAC0, DAC1};      //OUTPUT pins


int16_t analogOutputs[18] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
bool servoOutput[18] = {false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false};

int16_t timeData[7] = {0,0,0,0,0,0,0};


typedef union{
  float val;
  uint8_t bytes[4];
} floatValue;

struct Counter{
  uint8_t no;
  uint8_t reset;
  bool dn;
  bool dnNull;
  bool countNo;
  bool resetNo;
  bool countUp;
  bool countUpNull;
  uint8_t noIndex;
  uint8_t resetIndex;
  uint8_t dnIndex;
  uint8_t countUpIndex;
  uint8_t presetIndex;
  uint8_t countIndex;
};

struct CounterCTUD{
  uint8_t no;
  uint8_t noDown;
  uint8_t reset;
  bool dn;
  bool dnNull;
  bool countNo;
  bool countNoDown;
  bool resetNo;
  bool countUp;
  bool countDown;
  bool countUpNull;
  bool countDownNull;
  uint8_t noIndex;
  uint8_t noDownIndex;
  uint8_t resetIndex;
  uint8_t dnIndex;
  uint8_t countUpIndex;
  uint8_t countDownIndex;
  uint8_t presetIndex;
  uint8_t countIndex;
};


struct Timer{
  uint8_t no;
  bool dn;
  bool dnNull;
  uint8_t reset;
  bool tt;
  bool enableNo;
  bool activeTimer;
  bool ttNull;
  bool timerDone;
  int typeTimer;
  uint8_t noIndex;
  uint8_t dnIndex;
  uint8_t ttIndex;
  uint8_t resetIndex;
  int8_t presetIndex;
  int8_t currentTimeIndex;
  unsigned long startTime;
  bool isPoint1;
};


unsigned long startTime;
unsigned long delta;
unsigned long targetTime;

unsigned long PC; //HERE
uint8_t n;
int mask = (0x3FFF);
bool A;
bool X;
bool Y;

bool notFlag;
bool risingFlag;
bool fallingFlag;
bool iFlag;
bool iFlag1;
bool oFlag1;
bool oFlag;
bool vFlag;
bool vFlag1;
bool vFlag2;
bool vFlag3;
bool vFlag4;
bool vFlag5;
bool vFlag6;

bool latchFlag;
bool unlatchFlag;
bool negateFlag;
bool coilFlag;

bool timerFlag;
bool counterFlag;
bool counterCTUDFlag;

bool inFunction;

bool configStream = false;
bool enableHMIComm = false;
int8_t inputSize;
int8_t outputSize;
int8_t virtualSize;

struct Counter myCounter[24];
struct CounterCTUD myCounterCTUD[24];
struct Timer myTimer[24];
char bufferSerial[128];
int buffSize;
bool inCount;
int8_t waitReadCounter;
uint8_t ld;
uint8_t md;
uint8_t hd;
uint8_t ld32;
uint8_t md32;
uint8_t hd32;
uint16_t tempValue;
uint32_t tempValue32;
bool sendDataConfig = false;

int8_t iSize, oSize, vSize, eSize, dSize, floatSize;

Servo myServo0;
Servo myServo1;
Servo myServo2;
Servo myServo3;
Servo myServo4;
Servo myServo5;
Servo myServo6;
Servo myServo7;

void checkSpace();
void reprogram();
void initProgram();
void interpreter();
void updateRegisters();
bool setDnCounter(int posi, bool dnState);
bool setDnCounterCTUD(int posi, bool dnState);
void cleanOldCounters(int tempCountCounter);
void cleanOldTimers(int tempCountTimer);
void blinking();
void checkVariablesDint();
int readAnalogPin(int index);
void writeAnalogPin(int index, int value);
int readAnalogOutput(int index);
void setDnValue(bool newState, int tempPC);
bool getNoValue();
float getSource();
void setDest(float valueDest);
int16_t getJoinValue();
bool isTagEnable(int tempPC);

void checkVariablesAOut();
void checkVariablesInt();
void checkVariablesDint();
void checkVariablesReal();
void readRawVariables();
bool isSourceInt();
int getOutIndex();
bool getUpEdgeActive();
void updateVariable(int tempPC, int32_t valueDest);
void setDest(int tempPC, float valueDest);

void equ();
void neq();
void les();
void leq();
void grt();
void geq();
void mov();
void lim();
void meq();
void andOp();
void orOp();
void notOp();
void xorOp();
void addOp();
void subOp();
void mulOp();
void divOp();
void xpy();
void absMath();
void sqrMath();
void asnMath();
void acsMath();
void atnMath();
void cosMath();
void lnMath();
void logMath();
void sinMath();
void tanMath();
void scp();
void scl();

void bsl();
void bsr();
void ror();
void rol();

void setup() {

  Serial1.begin(9600);

  getTime();
  
  pinMode(13, OUTPUT);
  digitalWrite(13, LOW);


  for(int i=0; i<16; i++){       //pinMode for all the pins of the PLC UNO
    pinMode(iPins[i],INPUT);
    pinMode(oPins[i],OUTPUT);
  }

  updateTimeRegisters();


  startReprogramTime =  millis();
  startBlinkReproTime = millis();

  while(dueFlashStorage.read(REG_STATUS)!='2'){
    reprogram();
    delay(500);
  }

  runReprogramTime = enableHMImode;

  initProgram();
  
}

void loop() {
   
  
  // put your main code here, to run repeatedly:
  if(dueFlashStorage.read(REG_STATUS)=='2'){
    digitalWrite(13, HIGH);

    while(true){

      interpreter();
      updateRegisters();
      updateTimeRegisters();
      reprogram();
      updateTimeFromRTC();
     

    }//END while

  }//END IF dueFlashStorage.read

  
}//END loop()


void interpreter(){
  PC=0; //put the PC in the start of the memory
  int dataByte = dueFlashStorage.read(PC);
  
  while(dataByte!='z'){ //z is the value that indicate that the program has ended

    if(inFunction){
        switch(dataByte){
          case 'a':
            equ();
          break;
          case 'b':
            neq();
          break;
          case 'c':
            les();
          break;
          case 'd':
            leq();
          break;
          case 'e':
            grt();
          break;
          case 'f':
            geq();
          break;
          case 'g':
            lim();
          break;
          case 'h':
            meq();
          break;
          case 'i':
            mov();
          break;
          case 'j':
            bsl();
          break;
          case 'k':
            bsr();
          break;
          case 'l':
            rol();
          break;
          case 'm':
            ror();
          break;
          case 'n':
            andOp();
          break;
          case 'o':
            orOp();
          break;
          case 'p':
            notOp();
          break;
          case 'q':
            xorOp();
          break;
          case 'r':
            addOp();
          break;
          case 's':
            subOp();
          break;
          case 't':
            mulOp();
          break;
          case 'u':
            divOp();
          break;
          case 'v':
            absMath();
          break;
          case 'w':
            sqrMath();
          break;
          case 'x':
            xpy();
          break;
          case 'y':
            asnMath();
          break;
          case 'I':
            acsMath();
          break;
          case 'A':
            atnMath();
          break;
          case 'B':
            cosMath();
          break;
          case 'C':
            lnMath();
          break;
          case 'D':
            logMath();
          break;
          case 'E':
            sinMath();
          break;
          case 'F':
            tanMath();
          break;
          case 'G':
            scp();
          break;
          case 'H':
            scl();
          break;
          default:
            Serial.println("ERROR");
          break;

        }//END swith()
        inFunction = false;
    }

    else{
      switch(dataByte){
        case '$': //functions
          inFunction = true;
          break;
        case 'I': //i
          iFlag=true;
          break;
        case 'i':
          iFlag1 = true;
          break;
        case 'O': //o
          oFlag=true;
          break;
        case 'k':
          oFlag1 = true;
          break;
        case 'V': //v
          vFlag=true;
          break;
        case'v':
          vFlag1=true;
          break;
        case 'U': //v
          vFlag2=true;
          break;
        case'u':
          vFlag3=true;
          break;
        case 'A':
          vFlag4=true;
          break;
        case 'a':
          vFlag5=true;
          break;
        case 'c':
          vFlag6=true;
          break;
        case 'L': //Latch
          latchFlag=true;
          break;
        case 'l': //Unlatch
          unlatchFlag=true;
          break;
        case '|': //or
          Y= Y | A;
          A = true;
          break;
        case '}': //and
          A=Y | A;
          Y=false;
          break;
        case '~': //not
          notFlag=true;
          break;
        case 'e': //RISING EDGE
          risingFlag=true;
          break;
        case 'f': //FALLING EDGE
          fallingFlag=true;
          break;
        case 'C': //coil
          coilFlag=true;
          break;
        case 'n': //negate coil
          negateFlag = true;
          break;
        case 'R': //rung
          A=true;
          Y=false;
          coilFlag = false;
          break;
        case 'q': //timer
          timerFlag=true;
          coilFlag = false;
          break;
        case 'B': //counter CTU
          counterFlag=true;
          coilFlag = false;
          break;
        case 'b': //counter CTUD
          counterCTUDFlag = true;
          coilFlag = false;
          break;
        default:
          dataByte = dataByte - 48;
          if((dataByte<15)&&(dataByte>=0)){  //posible to be changed

            n=dataByte;

            if(coilFlag){
              if(iFlag){
                iState[n]=A;
                iFlag=false;
              }

              else if(oFlag){
                if(latchFlag){
                  latchFlag=false;
                  if(A) oState[n]=true;

                }//END if(latchFlag)
                else if(unlatchFlag){
                  unlatchFlag=false;
                  if(A) oState[n]=false;
                }

                else if(negateFlag){
                  negateFlag=false;
                  if(A) oState[n]=false;
                  else oState[n]=true;
                }

                else if((!latchFlag) && (!unlatchFlag) && (!negateFlag)){
                  oState[n]=A;
                }
                oFlag=false;
              }

              else if(oFlag1){
                if(latchFlag){
                  latchFlag=false;
                  if(A) oState[10+n]=true;

                }//END if(latchFlag)
                else if(unlatchFlag){
                  unlatchFlag=false;
                  if(A) oState[10+n]=false;
                }

                else if(negateFlag){
                  negateFlag=false;
                  if(A) oState[10+n]=false;
                  else oState[10+n]=true;
                }

                else if((!latchFlag) && (!unlatchFlag) && (!negateFlag)){
                  oState[10+n]=A;
                }
                oFlag1=false;
              }

              else if(vFlag){
                if(latchFlag){
                  latchFlag=false;
                  if(A) vState[n]=true;

                }//END if(latchFlag)
                else if(unlatchFlag){
                  unlatchFlag=false;
                  if(A) vState[n]=false;
                }

                else if(negateFlag){
                  negateFlag=false;
                  if(A) vState[n]=false;
                  else vState[n]=true;
                }

                else if((!latchFlag) && (!unlatchFlag) && (!negateFlag)){
                  vState[n]=A;
                }
                vFlag=false;
              }

              else if(vFlag1){
                if(latchFlag){
                  latchFlag=false;
                  if(A) vState[10+n]=true;

                }//END if(latchFlag)
                else if(unlatchFlag){
                  unlatchFlag=false;
                  if(A) vState[10+n]=false;
                }

                else if(negateFlag){
                  negateFlag=false;
                  if(A) vState[10+n]=false;
                  else vState[10+n]=true;
                }

                else if((!latchFlag) && (!unlatchFlag) && (!negateFlag)){
                  vState[10+n]=A;
                }
                vFlag1 = false;
              }

              else if(vFlag2){
                if(latchFlag){
                  latchFlag=false;
                  if(A) vState[20+n]=true;

                }//END if(latchFlag)
                else if(unlatchFlag){
                  unlatchFlag=false;
                  if(A) vState[20+n]=false;
                }

                else if(negateFlag){
                  negateFlag=false;
                  if(A) vState[20+n]=false;
                  else vState[20+n]=true;
                }

                else if((!latchFlag) && (!unlatchFlag) && (!negateFlag)){
                  vState[20+n]=A;
                }
                vFlag2 = false;
              }

              else if(vFlag3){
                if(latchFlag){
                  latchFlag=false;
                  if(A) vState[30+n]=true;

                }//END if(latchFlag)
                else if(unlatchFlag){
                  unlatchFlag=false;
                  if(A) vState[30+n]=false;
                }

                else if(negateFlag){
                  negateFlag=false;
                  if(A) vState[30+n]=false;
                  else vState[30+n]=true;
                }

                else if((!latchFlag) && (!unlatchFlag) && (!negateFlag)){
                  vState[30+n]=A;
                }
                vFlag3 = false;
              }

              else if(vFlag4){
                if(latchFlag){
                  latchFlag=false;
                  if(A) vState[40+n]=true;

                }//END if(latchFlag)
                else if(unlatchFlag){
                  unlatchFlag=false;
                  if(A) vState[40+n]=false;
                }

                else if(negateFlag){
                  negateFlag=false;
                  if(A) vState[40+n]=false;
                  else vState[40+n]=true;
                }

                else if((!latchFlag) && (!unlatchFlag) && (!negateFlag)){
                  vState[40+n]=A;
                }
                vFlag4 = false;
              }

              else if(vFlag5){
                if(latchFlag){
                  latchFlag=false;
                  if(A) vState[50+n]=true;

                }//END if(latchFlag)
                else if(unlatchFlag){
                  unlatchFlag=false;
                  if(A) vState[50+n]=false;
                }

                else if(negateFlag){
                  negateFlag=false;
                  if(A) vState[50+n]=false;
                  else vState[50+n]=true;
                }

                else if((!latchFlag) && (!unlatchFlag) && (!negateFlag)){
                  vState[50+n]=A;
                }
                vFlag5 = false;
              }

              else if(vFlag6){
                if(latchFlag){
                  latchFlag=false;
                  if(A) vState[60+n]=true;

                }//END if(latchFlag)
                else if(unlatchFlag){
                  unlatchFlag=false;
                  if(A) vState[60+n]=false;
                }

                else if(negateFlag){
                  negateFlag=false;
                  if(A) vState[60+n]=false;
                  else vState[60+n]=true;
                }

                else if((!latchFlag) && (!unlatchFlag) && (!negateFlag)){
                  vState[60+n]=A;
                }
                vFlag6 = false;
              }

              //coilFlag=false;
            }//END if(coilFlag)

            else if((!coilFlag)&&(!counterFlag)&&(!counterCTUDFlag)&&(!timerFlag)){

              if((iFlag)){
                if((notFlag) && (!risingFlag) && (!fallingFlag)){
                  if(iState[n]) X = false;
                  else  X = true;
                  //X = ~(iState[n]);
                  A = A & X;
                  notFlag=false;
                }//END IF (notFlag)
                else if((!notFlag) && (risingFlag) && (!fallingFlag)){
                  risingFlag = false;

                  if(iUpEdge[n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if((!notFlag) && (!risingFlag) && (fallingFlag)){
                  fallingFlag = false;

                  if(iDownEdge[n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if((!notFlag) && (!risingFlag) && (!fallingFlag)) {
                  A=(iState[n] & A);
                }
                iFlag=false;
              }//END if(iFlag)

              else if((iFlag1)){
                if(notFlag){
                  if(iState[10+n]) X = false;
                  else  X = true;
                  //X = ~(iState[10+n]);
                  A = A & X;
                  notFlag=false;
                }//END IF (notFlag)
                else if((!notFlag) && (risingFlag) && (!fallingFlag)){
                  risingFlag=false;

                  if(iUpEdge[10+n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if((!notFlag) && (!risingFlag) && (fallingFlag)){
                  fallingFlag=false;

                  if(iDownEdge[10+n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if((!notFlag) && (!risingFlag) && (!fallingFlag)) {
                  A=(iState[10+n] & A);
                }//END else
                iFlag1=false;
              }//END if(iFlag)

              else if((oFlag)){
                n=dataByte;
                if((notFlag) && (!risingFlag) && (!fallingFlag)){
                  if(oState[n]) X = false;
                  else  X = true;
                  //X = ~(oState[n]);
                  A = A & X;
                  notFlag=false;
                }//END IF (notFlag)
                else if((!notFlag)&&(risingFlag)&&(!fallingFlag)){
                  risingFlag = false;

                  if(oUpEdge[n]) X = true;
                  else X = false;
                  A = A & X;
                }
                else if((!notFlag)&&(!risingFlag)&&(fallingFlag)){
                  fallingFlag = false;

                  if(oDownEdge[n]) X = true;
                  else X = false;
                  A = A & X;
                }
                else if((!notFlag)&&(!risingFlag)&&(!fallingFlag)){
                  A=(oState[n] & A);
                }//END else
                oFlag=false;
              }//END if(oFlag)

              else if((oFlag1)){
                n=dataByte;
                if((notFlag) && (!risingFlag) && (!fallingFlag)){
                  if(oState[10+n]) X = false;
                  else  X = true;
                  //X = ~(oState[10+n]);
                  A = A & X;
                  notFlag=false;
                }//END IF (notFlag)
                else if((!notFlag)&&(risingFlag)&&(!fallingFlag)){
                  risingFlag = false;

                  if(oUpEdge[10+n]) X = true;
                  else X = false;
                  A = A & X;
                }
                else if((!notFlag)&&(!risingFlag)&&(fallingFlag)){
                  fallingFlag = false;

                  if(oDownEdge[10+n]) X = true;
                  else X = false;
                  A = A & X;
                }
                else if((!notFlag)&&(!risingFlag)&&(!fallingFlag)){
                  A=(oState[10+n] & A);
                }//END else
                oFlag1=false;
              }//END if(oFlag)

              else if((vFlag)){
                n=dataByte;
                if((notFlag) && (!risingFlag) && (!fallingFlag)){
                  if(vState[n]) X = false;
                  else  X = true;
                  A = A & X;
                  notFlag=false;
                }//END
                else if((!notFlag) && (risingFlag) && (!fallingFlag)){
                  risingFlag = false;

                  if(vUpEdge[n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if((!notFlag) && (!risingFlag) && (fallingFlag)){
                  fallingFlag = false;

                  if(vDownEdge[n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if((!notFlag) && (!risingFlag) && (!fallingFlag)) {
                  A=(vState[n] & A);
                }//END else
                vFlag=false;
              }//END if(vFlag)

              else if((vFlag1)){
                n=dataByte;
                if((notFlag) && (!risingFlag) && (!fallingFlag)){
                  if(vState[10+n]) X = false;
                  else  X = true;
                  A = A & X;
                  notFlag=false;
                }//END IF (notFlag)
                else if((!notFlag) && (risingFlag) && (!fallingFlag)){
                  risingFlag = false;

                  if(vUpEdge[10+n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if((!notFlag) && (!risingFlag) && (fallingFlag)){
                  fallingFlag = false;

                  if(vDownEdge[10+n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if((!notFlag) && (!risingFlag) && (!fallingFlag)) {
                  A=(vState[10+n] & A);
                }//END else
                vFlag1=false;
              }//END if(vFlag1)

              else if((vFlag2)){
                n=dataByte;
                if((notFlag) && (!risingFlag) && (!fallingFlag)){
                  if(vState[20+n]) X = false;
                  else  X = true;
                  A = A & X;
                  notFlag=false;
                }//END IF (notFlag)
                else if((!notFlag) && (risingFlag) && (!fallingFlag)){
                  risingFlag = false;

                  if(vUpEdge[20+n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if((!notFlag) && (!risingFlag) && (fallingFlag)){
                  fallingFlag = false;

                  if(vDownEdge[20+n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if((!notFlag) && (!risingFlag) && (!fallingFlag)) {
                  A=(vState[20+n] & A);
                }//END else
                vFlag2=false;
              }//END if(vFlag2)

              else if((vFlag3)){
                n=dataByte;
                if((notFlag) && (!risingFlag) && (!fallingFlag)){
                  if(vState[30+n]) X = false;
                  else  X = true;
                  A = A & X;
                  notFlag=false;
                }//END IF (notFlag)
                else if((!notFlag) && (risingFlag) && (!fallingFlag)){
                  risingFlag = false;

                  if(vUpEdge[30+n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if((!notFlag) && (!risingFlag) && (fallingFlag)){
                  fallingFlag = false;

                  if(vDownEdge[30+n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if((!notFlag) && (!risingFlag) && (!fallingFlag)) {
                  A=(vState[30+n] & A);
                }//END else
                vFlag3=false;
              }//END if(vFlag3)

              else if((vFlag4)){
                n=dataByte;
                if((notFlag) && (!risingFlag) && (!fallingFlag)){
                  if(vState[40+n]) X = false;
                  else  X = true;
                  A = A & X;
                  notFlag=false;
                }//END IF (notFlag)
                else if((!notFlag) && (risingFlag) && (!fallingFlag)){
                  risingFlag = false;

                  if(vUpEdge[40+n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if((!notFlag) && (!risingFlag) && (fallingFlag)){
                  fallingFlag = false;

                  if(vDownEdge[40+n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if((!notFlag) && (!risingFlag) && (!fallingFlag)) {
                  A=(vState[40+n] & A);
                }//END else
                vFlag4=false;
              }//END if(vFlag4)

              else if((vFlag5)){
                n=dataByte;
                if((notFlag) && (!risingFlag) && (!fallingFlag)){
                  if(vState[50+n]) X = false;
                  else  X = true;
                  A = A & X;
                  notFlag=false;
                }//END IF (notFlag)
                else if((!notFlag) && (risingFlag) && (!fallingFlag)){
                  risingFlag = false;

                  if(vUpEdge[50+n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if((!notFlag) && (!risingFlag) && (fallingFlag)){
                  fallingFlag = false;

                  if(vDownEdge[50+n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if((!notFlag) && (!risingFlag) && (!fallingFlag)) {
                  A=(vState[50+n] & A);
                }//END else
                vFlag5=false;
              }//END if(vFlag5)

              else if((vFlag6)){
                n=dataByte;
                if((notFlag) && (!risingFlag) && (!fallingFlag)){
                  if(vState[60+n]) X = false;
                  else  X = true;
                  A = A & X;
                  notFlag=false;
                }//END IF (notFlag)
                else if((!notFlag) && (risingFlag) && (!fallingFlag)){
                  risingFlag = false;

                  if(vUpEdge[60+n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if((!notFlag) && (!risingFlag) && (fallingFlag)){
                  fallingFlag = false;

                  if(vDownEdge[60+n]) X = true;
                  else  X = false;
                  A = A & X;
                }
                else if((!notFlag) && (!risingFlag) && (!fallingFlag)) {
                  A=(vState[60+n] & A);
                }//END else
                vFlag6=false;
              }//END if(vFlag6)

            }//END else if(~cFlag)


            else if((!coilFlag)&&(counterFlag)&&(!counterCTUDFlag)&&(!timerFlag)){    //Counter Section CTU


              bool tempNo;
              bool tempReset;
              bool tempDn;

              switch(myCounter[n].no){
                case 0:
                  tempNo = vState[myCounter[n].noIndex];
                break;

                case 1:
                  tempNo = iState[myCounter[n].noIndex];
                break;

                case 2:
                  tempNo = oState[myCounter[n].noIndex];
                break;
              }

              switch(myCounter[n].reset){
                case 0:
                  tempReset = vState[myCounter[n].resetIndex];
                break;

                case 1:
                  tempReset = iState[myCounter[n].resetIndex];
                break;

                case 2:
                  tempReset = oState[myCounter[n].resetIndex];
                break;
              }

              if(myCounter[n].dn) tempDn = oState[myCounter[n].dnIndex];
              else tempDn = vState[myCounter[n].dnIndex];


              //PUT the same code of the PLC Ladder
              

              if((tempNo)&&(!tempReset)){

                if(!myCounter[n].resetNo){
                  

                  if((tempNo)&&(!myCounter[n].countNo)){
                 
                      dintValues[myCounter[n].countIndex]++;

                      myCounter[n].countNo = true;
                      bool returned = setDnCounter(n,tempDn);
                      if(!myCounter[n].dnNull){
                        if(myCounter[n].dn) oState[myCounter[n].dnIndex] = returned;
                        else vState[myCounter[n].dnIndex] = returned;
                      }

                      if(!myCounter[n].countUpNull){
                        if(myCounter[n].countUp)  oState[myCounter[n].countUpIndex] = true;
                        else  vState[myCounter[n].countUpIndex] = true;
                      }



                  }
                }//END if(!myCounter[n].resetNo)
              }//END if((tempNo)&&(!tempReset))

              else if((!tempNo)&&(myCounter[n].countNo))  {
                myCounter[n].countNo = false;

                if(!myCounter[n].countUpNull){
                  if(myCounter[n].countUp)  oState[myCounter[n].countUpIndex] = false;
                  else  vState[myCounter[n].countUpIndex] = false;
                }
              }

              if((tempReset)&&(!myCounter[n].resetNo)){
                myCounter[n].resetNo = true;
                bool returned = setDnCounter(n,tempDn);
                if(!myCounter[n].dnNull){
                  if(myCounter[n].dn) oState[myCounter[n].dnIndex] = returned;
                  else vState[myCounter[n].dnIndex] = returned;
                }
                dintValues[myCounter[n].countIndex] = 0;

                if(!myCounter[n].countUpNull){
                  if(myCounter[n].countUp)  oState[myCounter[n].countUpIndex] = false;
                  else  vState[myCounter[n].countUpIndex] = false;
                }
              }

              else if((!tempReset)&&(myCounter[n].resetNo)) {
                myCounter[n].resetNo = false;

              }


              counterFlag=false;


            }//END Counter Section

            else if((!coilFlag)&&(!counterFlag)&&(counterCTUDFlag)&&(!timerFlag)){    //Counter Section CTUD

              bool tempNo;
              bool tempReset;
              bool tempNoDown;
              bool tempDn;

              switch(myCounterCTUD[n].no){
                case 0:
                  tempNo = vState[myCounterCTUD[n].noIndex];
                break;

                case 1:
                  tempNo = iState[myCounterCTUD[n].noIndex];
                break;

                case 2:
                  tempNo = oState[myCounterCTUD[n].noIndex];
                break;
              }

             switch(myCounterCTUD[n].noDown){
                case 0:
                  tempNoDown = vState[myCounterCTUD[n].noDownIndex];
                break;

                case 1:
                  tempNoDown = iState[myCounterCTUD[n].noDownIndex];
                break;

                case 2:
                  tempNoDown = oState[myCounterCTUD[n].noDownIndex];
                break;
             }

              switch(myCounterCTUD[n].reset){
                case 0:
                  tempReset = vState[myCounterCTUD[n].resetIndex];
                break;

                case 1:
                  tempReset = iState[myCounterCTUD[n].resetIndex];
                break;

                case 2:
                  tempReset = oState[myCounterCTUD[n].resetIndex];
                break;
             }


              if(myCounterCTUD[n].dn) tempDn = oState[myCounterCTUD[n].dnIndex];
              else tempDn = vState[myCounterCTUD[n].dnIndex];


              //PUT the same code of the PLC Ladder

              if((tempNo)&&(!tempReset)){
                if(!myCounterCTUD[n].resetNo){
                  if((tempNo)&&(!myCounterCTUD[n].countNo)){

                      dintValues[myCounterCTUD[n].countIndex]++;
                      myCounterCTUD[n].countNo = true;
                      bool returned = setDnCounterCTUD(n,tempDn);
                      if(!myCounterCTUD[n].dnNull){
                        if(myCounterCTUD[n].dn) oState[myCounterCTUD[n].dnIndex] = returned;
                        else vState[myCounterCTUD[n].dnIndex] = returned;
                      }

                      if(!myCounterCTUD[n].countUpNull){
                        if(myCounterCTUD[n].countUp)  oState[myCounterCTUD[n].countUpIndex] = true;
                        else  vState[myCounterCTUD[n].countUpIndex] = true;
                      }

                  }
                }//END if(!myCounterCTUD[n].resetNo)
              }//END if((tempNo)&&(!tempReset))

              else if((!tempNo)&&(myCounterCTUD[n].countNo))  {
                myCounterCTUD[n].countNo = false;

                if(!myCounterCTUD[n].countUpNull){
                  if(myCounterCTUD[n].countUp)  oState[myCounterCTUD[n].countUpIndex] = false;
                  else  vState[myCounterCTUD[n].countUpIndex] = false;
                }
              }//END else if((!tempNo) && (myCounterCTUD[n].countNo)


              else if((tempNoDown) && (!tempReset)){
                if(!myCounterCTUD[n].resetNo){
                  if(!myCounterCTUD[n].countNoDown){

                      dintValues[myCounterCTUD[n].countIndex]--;
                      myCounterCTUD[n].countNoDown = true;
                      bool returned = setDnCounterCTUD(n,tempDn);
                      if(!myCounterCTUD[n].dnNull){
                        if(myCounterCTUD[n].dn) oState[myCounterCTUD[n].dnIndex] = returned;
                        else vState[myCounterCTUD[n].dnIndex] = returned;
                      }

                      if(!myCounterCTUD[n].countDownNull){
                        if(myCounterCTUD[n].countDown)  oState[myCounterCTUD[n].countDownIndex] = true;
                        else  vState[myCounterCTUD[n].countDownIndex] = true;
                      }

                  }//END if(!myCounterCTUD[n].countNoDown)
                }//END if(!myCounterCTUD[n].resetNo)
              }//END else if((tempNoDown) && (!tempReset))


              else if((!tempNoDown) && (myCounterCTUD[n].countNoDown)){
                myCounterCTUD[n].countNoDown = false;

                if(!myCounterCTUD[n].countDownNull){
                  if(myCounterCTUD[n].countDown)  oState[myCounterCTUD[n].countDownIndex] = false;
                  else  vState[myCounterCTUD[n].countDownIndex] = false;
                }
              }//END else if((!tempNoDown) && (myCounterCTUD[n].countNoDown))

              if((tempReset)&&(!myCounterCTUD[n].resetNo)){
                myCounterCTUD[n].resetNo = true;
                bool returned = setDnCounterCTUD(n,tempDn);
                if(!myCounterCTUD[n].dnNull){
                  if(myCounterCTUD[n].dn) oState[myCounterCTUD[n].dnIndex] = returned;
                  else vState[myCounterCTUD[n].dnIndex] = returned;
                }
                dintValues[myCounterCTUD[n].countIndex] = 0;

                if(!myCounterCTUD[n].countUpNull){
                  if(myCounterCTUD[n].countUp)  oState[myCounterCTUD[n].countUpIndex] = false;
                  else  vState[myCounterCTUD[n].countUpIndex] = false;
                }
              }

              else if((!tempReset)&&(myCounterCTUD[n].resetNo)) {
                myCounterCTUD[n].resetNo = false;

              }


              counterCTUDFlag=false;


            }//END Counter Section



            else if((!coilFlag)&&(!counterFlag)&&(!counterCTUDFlag)&&(timerFlag)){    //Timer Section

              bool tempNo;

              unsigned long tempTime;

              switch(myTimer[n].no){
                case 0:
                  tempNo = vState[myTimer[n].noIndex];
                break;

                case 1:
                  tempNo = iState[myTimer[n].noIndex];
                break;

                case 2:
                  tempNo = oState[myTimer[n].noIndex];
                break;
              }


              switch(myTimer[n].typeTimer){
                case 1:                                                           //TON
                    if(tempNo){
                      if(!myTimer[n].activeTimer){
                        myTimer[n].activeTimer = true;

                        if(!myTimer[n].ttNull){
                          if(myTimer[n].tt) oState[myTimer[n].ttIndex] = true;
                          else vState[myTimer[n].ttIndex] = true;
                        }

                        myTimer[n].startTime = millis();
                      }
                    }

                    else{
                      myTimer[n].activeTimer = false;


                      if(!myTimer[n].ttNull){
                        if(myTimer[n].tt) oState[myTimer[n].ttIndex] = false;
                        else vState[myTimer[n].ttIndex] = false;
                      }

                      if(!myTimer[n].dnNull){

                        if(myTimer[n].dn) oState[myTimer[n].dnIndex] = false;
                        else vState[myTimer[n].dnIndex] = false;

                        dintValues[myTimer[n].currentTimeIndex] = 0;

                      }
                    }

                    if(myTimer[n].activeTimer){
                      tempTime = millis();
                      tempTime = tempTime - myTimer[n].startTime;

                      if(tempTime < 0){                                                                       
                        myTimer[n].startTime = millis();
                        tempTime = millis();
                        tempTime = tempTime - myTimer[n].startTime;
                      }

                      dintValues[myTimer[n].currentTimeIndex] = (int32_t)(tempTime*(0.001f));

                      uint32_t tempPreset = 0;
                      if(myTimer[n].isPoint1)
                        tempPreset = (uint32_t)(((0.1)*(dintValues[myTimer[n].presetIndex]))*(1000L));
                      else
                        tempPreset = (uint32_t)((dintValues[myTimer[n].presetIndex])*(1000L));

                      if(tempTime >= tempPreset){
                        delay(20);

                        if(!myTimer[n].ttNull){
                          if(myTimer[n].tt) oState[myTimer[n].ttIndex] = false;
                          else vState[myTimer[n].ttIndex] = false;
                        }

                        if(!myTimer[n].dnNull){
                          if(myTimer[n].dn) oState[myTimer[n].dnIndex] = true;
                          else vState[myTimer[n].dnIndex] = true;
                        }
                      }
                    }
                  break;

                case 2:                                                                 //TOF
                        if(tempNo){
                          if(!myTimer[n].enableNo){
                            myTimer[n].enableNo = true;

                            if(!myTimer[n].ttNull){
                              if(myTimer[n].tt) oState[myTimer[n].ttIndex] = false;
                              else vState[myTimer[n].ttIndex] = false;
                            }

                            if(!myTimer[n].dnNull){
                              if(myTimer[n].dn) oState[myTimer[n].dnIndex] = true;
                              else vState[myTimer[n].dnIndex] = true;
                            }
                          }//END if(!myTimer[n].enableNo)

                          else  {
                            myTimer[n].activeTimer = false;
                            dintValues[myTimer[n].currentTimeIndex] = 0;
                          }
                        }//END if(tempNo

                        else{
                          if(myTimer[n].enableNo){
                            myTimer[n].enableNo = false;

                            if(!myTimer[n].ttNull){
                              if(myTimer[n].tt) oState[myTimer[n].ttIndex] = true;
                              else vState[myTimer[n].ttIndex] = true;
                            }

                            myTimer[n].activeTimer = true;
                            myTimer[n].startTime = millis();
                          }//END if(myTimer[n].enableNo)
                        }//END else

                        if(myTimer[n].activeTimer){
                          tempTime = millis();
                          tempTime = tempTime -myTimer[n].startTime;

                          if(tempTime < 0){                                                                       
                            myTimer[n].startTime = millis();
                            tempTime = millis();
                            tempTime = tempTime - myTimer[n].startTime;
                          }

                          dintValues[myTimer[n].currentTimeIndex] = (int32_t)(tempTime*(0.001f));

                          uint32_t tempPreset = 0;
                          if(myTimer[n].isPoint1)
                            tempPreset = (uint32_t)(((0.1)*(dintValues[myTimer[n].presetIndex]))*(1000L));
                          else
                            tempPreset = (uint32_t)((dintValues[myTimer[n].presetIndex])*(1000L));

                          if(tempTime >= tempPreset){
                            dintValues[myTimer[n].currentTimeIndex] = 0;
                            delay(20);
                            if(!myTimer[n].ttNull){
                              if(myTimer[n].tt) oState[myTimer[n].ttIndex] = false;
                              else vState[myTimer[n].ttIndex] = false;
                            }

                            if(!myTimer[n].dnNull){
                              if(myTimer[n].dn) oState[myTimer[n].dnIndex] = false;
                              else vState[myTimer[n].dnIndex] = false;
                            }

                            myTimer[n].activeTimer = false;
                          }//END if(tempTime >= myTimer[n].preset)

                        }//END if(myTimer[n].activeTimer)
                  break;


                  case 3:                                                                 //RTO

                    bool tempReset;

                    switch(myTimer[n].reset){
                      case 0:
                        tempReset = vState[myTimer[n].resetIndex];
                      break;

                      case 1:
                        tempReset = iState[myTimer[n].resetIndex];
                      break;

                      case 2:
                        tempReset = oState[myTimer[n].resetIndex];
                      break;
                    }


                    if(tempNo){
                      if((!myTimer[n].activeTimer)&&(!myTimer[n].timerDone)){
                        myTimer[n].activeTimer = true;

                        if(!myTimer[n].ttNull){
                          if(myTimer[n].tt) oState[myTimer[n].ttIndex] = true;
                          else vState[myTimer[n].ttIndex] = true;
                        }
                        myTimer[n].startTime = millis();
                      }
                    }//END if(tempNo)

                    else{
                      if(myTimer[n].activeTimer){
                        myTimer[n].activeTimer = false;

                        if(!myTimer[n].ttNull){
                          if(myTimer[n].tt) oState[myTimer[n].ttIndex] = false;
                          else vState[myTimer[n].ttIndex] = false;
                        }
                        long tempTime = millis();
                        int32_t tempValue = dintValues[myTimer[n].currentTimeIndex]*(1000L);
                        tempValue = tempValue + tempTime - myTimer[n].startTime;

                        int32_t tempStartCheck = tempTime - myTimer[n].startTime;
                        if(tempStartCheck < 0){                                                                       
                            myTimer[n].startTime = millis();
                            tempTime = millis();
                            tempValue = tempValue + tempTime - myTimer[n].startTime;
                        }
                      
                        
                        dintValues[myTimer[n].currentTimeIndex] = (int32_t)tempValue*(0.001f);

                      }//END if(mytimer[n].activeTimer)


                    }//END else

                    if(tempReset){
                      myTimer[n].activeTimer = false;
                      myTimer[n].enableNo = false;

                      if(!myTimer[n].ttNull){
                          if(myTimer[n].tt) oState[myTimer[n].ttIndex] = false;
                          else vState[myTimer[n].ttIndex] = false;
                      }

                      if(!myTimer[n].dnNull){
                        if(myTimer[n].dn) oState[myTimer[n].dnIndex] = false;
                        else vState[myTimer[n].dnIndex] = false;
                      }

                      dintValues[myTimer[n].currentTimeIndex] = 0;
                      myTimer[n].timerDone = false;
                    }


                    if(myTimer[n].activeTimer){
                      tempTime = millis();
                      tempTime = tempTime - myTimer[n].startTime;
                      int32_t tempValue = dintValues[myTimer[n].currentTimeIndex]*(1000L);
                      tempTime += tempValue;

                      uint32_t tempPreset = 0;
                      if(myTimer[n].isPoint1)
                        tempPreset = (uint32_t)(((0.1)*(dintValues[myTimer[n].presetIndex]))*(1000L));
                      else
                        tempPreset = (uint32_t)((dintValues[myTimer[n].presetIndex])*(1000L));

                      if(tempTime >= tempPreset){

                        delay(20);
                        if(!myTimer[n].ttNull){
                          if(myTimer[n].tt) oState[myTimer[n].ttIndex] = false;
                          else vState[myTimer[n].ttIndex] = false;
                        }//END if(!myTimer[n].ttNull)

                        if(!myTimer[n].dnNull){
                          if(myTimer[n].dn) oState[myTimer[n].dnIndex] = true;
                          else vState[myTimer[n].dnIndex] = true;
                        }

                        myTimer[n].timerDone = true;

                      }//END if(tempTime >= myTimer[n].preset)
                    }//END if(myTimer[n].activeTimer)

                  break;


              }//END switch(myTimer[n].typeTimer)

              timerFlag = false;
            }//END Timer Section

          }//END if(dataByte<15)
      }//END Switch
    }//END ELSE !INFUNCTION

    PC++;
    dataByte = dueFlashStorage.read(PC);

  }//END while(dataByte!=254)

  //END OF Z
  //Serial.println(" ");

}//END interpreter()

void updateRegisters(){
  
  for(int i = 0; i<16; i++){
    if(iMask[i]){           //input

      if((!iPreState[i]) && (iState[i]))
        iUpEdge[i] = true;
      else if((iPreState[i]) && (!iState[i]))
        iDownEdge[i] = true;
      else{
        iUpEdge[i] = false;
        iDownEdge[i] = false;
      }

      iPreState[i] = iState[i];
      iState[i]=digitalRead(iPins[i]);
      
    }

    if(oMask[i]){           //output

      if((!oPreState[i]) && (oState[i]))
        oUpEdge[i] = true;
      else if((oPreState[i]) && (!oState[i]))
        oDownEdge[i] = true;
      else{
        oUpEdge[i] = false;
        oDownEdge[i] = false;
      }

      oPreState[i] = oState[i];
      if(analogOutputs[i] == 0)
        digitalWrite(oPins[i], oState[i]);
    }
  }//END FOR input

  for(int i = 0; i<64; i++){
    if(vMask[i]){

      if((!vPreState[i]) && (vState[i]))
        vUpEdge[i] = true;
      else if((vPreState[i]) && (!vState[i]))
        vDownEdge[i] = true;
      else{
        vUpEdge[i] = false;
        vDownEdge[i] = false;
      }

      vPreState[i] = vState[i];
    }
  }
}//END updateRegisters()

  void cleanOutputs(){
      for(int i=0; i<16; i++)
        digitalWrite(oPins[i], false);
  }

  //fOR THE EEPROM 255 = true, 0 = false
  void initProgram(){
    if(dueFlashStorage.read(REG_STATUS)=='2'){   //memory status flag

      cleanOutputs();

      digitalWrite(13, HIGH);

      inputSize = 0;
      outputSize = 0;
      for(int i = 0; i<16; i++){   // i/o variables mask
            int tempVar = dueFlashStorage.read(REG_IN+i);//input
            
            if(tempVar == 126){
              iMask[i] = true;
              inputSize++;
            }

            tempVar = dueFlashStorage.read(REG_OUT+i); //output
            if(tempVar == 126){
              oMask[i] = true;
              outputSize++;
            }
      }//END FOR i/o mask

      //CHECK DAC PINS
      for(int i = 0; i<2; i++){
        int tempVar = dueFlashStorage.read(REG_OUT+i+16);//input
            if(tempVar == 126){
              dacAnalogMask[i] = true;
            }
      }

      virtualSize = 0;
      for(int i = 0; i<64; i++){   //virtual variables mask
            int tempVar = dueFlashStorage.read(REG_VIRTUAL+i);
            if(tempVar == 126){
              vMask[i] = true;
              virtualSize++;
            }
      }//END FOR virtual mask

      //INIT INT VARIABLES
      int tempPC = REG_INT;
      int32_t tempA = 0;
      int32_t tempB = 0;
      int intIndex = 0;

      while(tempPC<(REG_LONG - 1)){
        tempA = (int32_t)dueFlashStorage.read(tempPC);


        tempA = (int32_t)tempA << 8;
        tempPC++;

        tempA = tempA | (int32_t)dueFlashStorage.read(tempPC);
        tempPC++;

        intValues[intIndex] = (int16_t)tempA;

        intIndex++;

      }

      //INIT DINT VARIABLES
      tempPC = REG_LONG;
      tempA = 0;
      tempB = 0;
      intIndex = 0;

      while(tempPC < (REG_FLOAT-1)){
        tempA = dueFlashStorage.read(tempPC);
        tempA = ((int32_t)tempA) << 24;
        tempPC++;

        tempB = dueFlashStorage.read(tempPC);
        tempB = ((int32_t)tempB) << 16;
        tempA = tempA | tempB;
        tempPC++;

        tempB = dueFlashStorage.read(tempPC);
        tempB = ((int32_t)tempB) << 8;
        tempA = tempA |tempB;
        tempPC++;

        tempA = tempA | ((int32_t)dueFlashStorage.read(tempPC));
        tempPC++;

        dintValues[intIndex] = tempA;
        intIndex++;

      }

      //INIT REAL VARIABLES
      tempPC = REG_FLOAT;
      tempA = 0;
      tempB = 0;
      intIndex = 0;

      floatValue tempFloat;
      while(tempPC < (REG_TIMER-1)){
        tempFloat.bytes[3] = dueFlashStorage.read(tempPC);
        tempPC++;

        tempFloat.bytes[2] = dueFlashStorage.read(tempPC);
        tempPC++;

        tempFloat.bytes[1] = dueFlashStorage.read(tempPC);
        tempPC++;

        tempFloat.bytes[0] = dueFlashStorage.read(tempPC);
        tempPC++;

        realValues[intIndex] = tempFloat.val;

        intIndex++;

      }



      for(int i = 0; i<24; i++){   //Loading Counters 16 COUNTERS
        int tempMulti = REG_COUNTER + (16*i);
        char tempCounterValue = dueFlashStorage.read(tempMulti);

        if(tempCounterValue == 'B'){  //IF THE COUNTER IS ACTIVE  //CTU

          int tempFlag = dueFlashStorage.read(tempMulti+1);
          if(tempFlag == 'I') {
            myCounter[i].no = 1;
            myCounter[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48;
          }
          else if(tempFlag == 'i'){
            myCounter[i].no = 1;
            myCounter[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+10;
          }
          else if(tempFlag == 'O'){
            myCounter[i].no = 2;
            myCounter[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48;
          }
          else if(tempFlag == 'k'){
            myCounter[i].no = 2;
            myCounter[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+10;
          }
          else{
            myCounter[i].no = 0;
            switch(tempFlag){
              case 'V':
                  myCounter[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48;
                break;

              case 'v':
                  myCounter[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+10;
                break;

              case 'U':
                  myCounter[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+20;
                break;

              case 'u':
                  myCounter[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+30;
                break;

              case 'A':
                  myCounter[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+40;
                break;

              case 'a':
                  myCounter[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+50;
                break;

              case 'c':
                  myCounter[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+60;
                break;
            }//END switch
          }


          tempFlag = dueFlashStorage.read(tempMulti+3);
          if(tempFlag == 'I') {
            myCounter[i].reset = 1;
            myCounter[i].resetIndex = (dueFlashStorage.read(tempMulti+4))-48;
          }
          else if(tempFlag == 'i'){
            myCounter[i].reset = 1;
            myCounter[i].resetIndex = (dueFlashStorage.read(tempMulti+4))-48+10;
          }
          else if(tempFlag == 'O'){
            myCounter[i].reset = 2;
            myCounter[i].resetIndex = (dueFlashStorage.read(tempMulti+4))-48;
          }
          else if(tempFlag == 'k'){
            myCounter[i].reset = 2;
            myCounter[i].resetIndex = (dueFlashStorage.read(tempMulti+4))-48+10;
          }
          else {
            myCounter[i].reset = 0;
            switch(tempFlag){
              case 'V':
                  myCounter[i].resetIndex = (dueFlashStorage.read(tempMulti+4))-48;
                break;

              case 'v':
                  myCounter[i].resetIndex = (dueFlashStorage.read(tempMulti+4))-48+10;
                break;

              case 'U':
                  myCounter[i].resetIndex = (dueFlashStorage.read(tempMulti+4))-48+20;
                break;

              case 'u':
                  myCounter[i].resetIndex = (dueFlashStorage.read(tempMulti+4))-48+30;
                break;

              case 'A':
                  myCounter[i].resetIndex = (dueFlashStorage.read(tempMulti+4))-48+40;
                break;

              case 'a':
                  myCounter[i].resetIndex = (dueFlashStorage.read(tempMulti+4))-48+50;
                break;

              case 'c':
                  myCounter[i].resetIndex = (dueFlashStorage.read(tempMulti+4))-48+60;
                break;
            }//END switch
          }


          tempFlag = dueFlashStorage.read(tempMulti+5);
          if(tempFlag == 's'){ //null
            myCounter[i].dnNull = true;
          }
          else{
            myCounter[i].dnNull = false;
            if(tempFlag == 'O') {
              myCounter[i].dn = true;
              myCounter[i].dnIndex = (dueFlashStorage.read(tempMulti+6))-48;
            }
            else if(tempFlag == 'k'){
              myCounter[i].dn = true;
              myCounter[i].dnIndex = (dueFlashStorage.read(tempMulti+6))-48+10;
            }
            else {
              myCounter[i].dn = false;
              switch(tempFlag){
                case 'V':
                    myCounter[i].dnIndex = (dueFlashStorage.read(tempMulti+6))-48;
                  break;

                case 'v':
                    myCounter[i].dnIndex = (dueFlashStorage.read(tempMulti+6))-48+10;
                  break;

                case 'U':
                    myCounter[i].dnIndex = (dueFlashStorage.read(tempMulti+6))-48+20;
                  break;

                case 'u':
                    myCounter[i].dnIndex = (dueFlashStorage.read(tempMulti+6))-48+30;
                  break;

                case 'A':
                  myCounter[i].dnIndex = (dueFlashStorage.read(tempMulti+6))-48+40;
                break;

                case 'a':
                  myCounter[i].dnIndex = (dueFlashStorage.read(tempMulti+6))-48+50;
                break;

                case 'c':
                  myCounter[i].dnIndex = (dueFlashStorage.read(tempMulti+6))-48+60;
                break;

              }//END switch
            }
          }//END else

          tempFlag = dueFlashStorage.read(tempMulti+7);
          if(tempFlag == 's'){
            myCounter[i].countUpNull = true;
          }
          else{
            myCounter[i].countUpNull = false;
            if(tempFlag == 'O') {
              myCounter[i].countUp = true;
              myCounter[i].countUpIndex = (dueFlashStorage.read(tempMulti+8))-48;
            }
            else if(tempFlag == 'k'){
              myCounter[i].countUp = true;
              myCounter[i].countUpIndex = (dueFlashStorage.read(tempMulti+8))-48+10;
            }
            else {
              myCounter[i].countUp = false;
              switch(tempFlag){
                case 'V':
                    myCounter[i].countUpIndex = (dueFlashStorage.read(tempMulti+8))-48;
                  break;

                case 'v':
                    myCounter[i].countUpIndex = (dueFlashStorage.read(tempMulti+8))-48+10;
                  break;

                case 'U':
                    myCounter[i].countUpIndex = (dueFlashStorage.read(tempMulti+8))-48+20;
                  break;

                case 'u':
                    myCounter[i].countUpIndex = (dueFlashStorage.read(tempMulti+8))-48+30;
                  break;

                case 'A':
                  myCounter[i].countUpIndex = (dueFlashStorage.read(tempMulti+8))-48+40;
                break;

                case 'a':
                  myCounter[i].countUpIndex = (dueFlashStorage.read(tempMulti+8))-48+50;
                break;

                case 'c':
                  myCounter[i].countUpIndex = (dueFlashStorage.read(tempMulti+8))-48+60;
                break;
              }//END switch
            }
          }//END else


          tempFlag = dueFlashStorage.read(tempMulti+9);
          if(tempFlag == 'P'){
            myCounter[i].presetIndex = dueFlashStorage.read(tempMulti+10);
            myCounter[i].countIndex = dueFlashStorage.read(tempMulti+11);
            myCounter[i].countNo = false;
            myCounter[i].resetNo = false;
          }


        }//END if(244)


        else if(tempCounterValue == 'b'){                                               //CTUD

          int tempFlag = dueFlashStorage.read(tempMulti+1);
          if(tempFlag == 'I') {
            myCounterCTUD[i].no = 1;
            myCounterCTUD[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48;
          }
          else if(tempFlag == 'i'){
            myCounterCTUD[i].no = 1;
            myCounterCTUD[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+10;
          }
          else if(tempFlag == 'O'){
            myCounterCTUD[i].no = 2;
            myCounterCTUD[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48;
          }
          else if(tempFlag == 'k'){
            myCounterCTUD[i].no = 2;
            myCounterCTUD[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+10;
          }
          else{
            myCounterCTUD[i].no = 0;
            switch(tempFlag){
              case 'V':
                  myCounterCTUD[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48;
                break;

              case 'v':
                  myCounterCTUD[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+10;
                break;

              case 'U':
                  myCounterCTUD[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+20;
                break;

              case 'u':
                  myCounterCTUD[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+30;
                break;

             case 'A':
                  myCounterCTUD[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+40;
                break;

              case 'a':
                  myCounterCTUD[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+50;
                 break;

              case 'c':
                  myCounterCTUD[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+60;
                break;
            }//END switch
          }


          tempFlag = dueFlashStorage.read(tempMulti+3);
          if(tempFlag == 'I') {
            myCounterCTUD[i].reset = 1;
            myCounterCTUD[i].resetIndex = (dueFlashStorage.read(tempMulti+4))-48;
          }
          else if(tempFlag == 'i'){
            myCounterCTUD[i].reset = 1;
            myCounterCTUD[i].resetIndex = (dueFlashStorage.read(tempMulti+4))-48+10;
          }
          else if(tempFlag == 'O'){
            myCounterCTUD[i].reset = 2;
            myCounterCTUD[i].resetIndex = (dueFlashStorage.read(tempMulti+4))-48;
          }
          else if(tempFlag == 'k'){
            myCounterCTUD[i].reset = 2;
            myCounterCTUD[i].resetIndex = (dueFlashStorage.read(tempMulti+4))-48+10;
          }
          else {
            myCounterCTUD[i].reset = 0;
            switch(tempFlag){
              case 'V':
                  myCounterCTUD[i].resetIndex = (dueFlashStorage.read(tempMulti+4))-48;
                break;

              case 'v':
                  myCounterCTUD[i].resetIndex = (dueFlashStorage.read(tempMulti+4))-48+10;
                break;

              case 'U':
                  myCounterCTUD[i].resetIndex = (dueFlashStorage.read(tempMulti+4))-48+20;
                break;

              case 'u':
                  myCounterCTUD[i].resetIndex = (dueFlashStorage.read(tempMulti+4))-48+30;
                break;

              case 'A':
                  myCounterCTUD[i].resetIndex = (dueFlashStorage.read(tempMulti+4))-48+40;
                break;

              case 'a':
                  myCounterCTUD[i].resetIndex = (dueFlashStorage.read(tempMulti+4))-48+50;
                 break;

              case 'c':
                  myCounterCTUD[i].resetIndex = (dueFlashStorage.read(tempMulti+4))-48+60;
                break;
            }//END switch
          }


          tempFlag = dueFlashStorage.read(tempMulti+5);
          if(tempFlag == 's'){
            myCounterCTUD[i].dnNull = true;
          }
          else{
            myCounterCTUD[i].dnNull = false;
            if(tempFlag == 'O') {
              myCounterCTUD[i].dn = true;
              myCounterCTUD[i].dnIndex = (dueFlashStorage.read(tempMulti+6))-48;
            }
            else if(tempFlag == 'k'){
              myCounterCTUD[i].dn = true;
              myCounterCTUD[i].dnIndex = (dueFlashStorage.read(tempMulti+6))-48+10;
            }
            else {
              myCounterCTUD[i].dn = false;
              switch(tempFlag){
                case 'V':
                    myCounterCTUD[i].dnIndex = (dueFlashStorage.read(tempMulti+6))-48;
                  break;

                case 'v':
                    myCounterCTUD[i].dnIndex = (dueFlashStorage.read(tempMulti+6))-48+10;
                  break;

                case 'U':
                    myCounterCTUD[i].dnIndex = (dueFlashStorage.read(tempMulti+6))-48+20;
                  break;

                case 'u':
                    myCounterCTUD[i].dnIndex = (dueFlashStorage.read(tempMulti+6))-48+30;
                  break;

                case 'A':
                  myCounterCTUD[i].dnIndex = (dueFlashStorage.read(tempMulti+6))-48+40;
                break;

              case 'a':
                  myCounterCTUD[i].dnIndex = (dueFlashStorage.read(tempMulti+6))-48+50;
                 break;

              case 'c':
                  myCounterCTUD[i].dnIndex = (dueFlashStorage.read(tempMulti+6))-48+60;
                break;
              }//END switch
            }
          }//END else


          tempFlag = dueFlashStorage.read(tempMulti+7);
          if(tempFlag == 's'){
            myCounterCTUD[i].countUpNull = true;
          }
          else{
            myCounterCTUD[i].countUpNull = false;
            if(tempFlag == 'O') {
              myCounterCTUD[i].countUp = true;
              myCounterCTUD[i].countUpIndex = (dueFlashStorage.read(tempMulti+8))-48;
            }
            else if(tempFlag == 'k'){
              myCounterCTUD[i].countUp = true;
              myCounterCTUD[i].countUpIndex = (dueFlashStorage.read(tempMulti+8))-48+10;
            }
            else {
              myCounterCTUD[i].countUp = false;
              switch(tempFlag){
                case 'V':
                    myCounterCTUD[i].countUpIndex = (dueFlashStorage.read(tempMulti+8))-48;
                  break;

                case 'v':
                    myCounterCTUD[i].countUpIndex = (dueFlashStorage.read(tempMulti+8))-48+10;
                  break;

                case 'U':
                    myCounterCTUD[i].countUpIndex = (dueFlashStorage.read(tempMulti+8))-48+20;
                  break;

                case 'u':
                    myCounterCTUD[i].countUpIndex = (dueFlashStorage.read(tempMulti+8))-48+30;
                  break;

                  case 'A':
                  myCounterCTUD[i].countUpIndex = (dueFlashStorage.read(tempMulti+8))-48+40;
                   break;

                 case 'a':
                  myCounterCTUD[i].countUpIndex = (dueFlashStorage.read(tempMulti+8))-48+50;
                 break;

                 case 'c':
                  myCounterCTUD[i].countUpIndex = (dueFlashStorage.read(tempMulti+8))-48+60;
                break;
              }//END switch
            }
          }//END else


          tempFlag = dueFlashStorage.read(tempMulti+9);
          if(tempFlag == 'P'){

            myCounterCTUD[i].presetIndex = dueFlashStorage.read(tempMulti+10);
            myCounterCTUD[i].countIndex = dueFlashStorage.read(tempMulti+11);
            myCounterCTUD[i].countNo = false;
            myCounterCTUD[i].resetNo = false;
          }


          tempFlag = dueFlashStorage.read(tempMulti+12);                                         //N.O. DOWN
          if(tempFlag == 'I') {
            myCounterCTUD[i].noDown = 1;
            myCounterCTUD[i].noDownIndex = (dueFlashStorage.read(tempMulti+13))-48;
          }
          else if(tempFlag == 'i'){
            myCounterCTUD[i].noDown = 1;
            myCounterCTUD[i].noDownIndex = (dueFlashStorage.read(tempMulti+13))-48+10;
          }
          else if(tempFlag == 'O'){
            myCounterCTUD[i].noDown = 2;
            myCounterCTUD[i].noDownIndex = (dueFlashStorage.read(tempMulti+13))-48;
          }
          else if(tempFlag == 'k'){
            myCounterCTUD[i].noDown = 2;
            myCounterCTUD[i].noDownIndex = (dueFlashStorage.read(tempMulti+13))-48+10;
          }
          else{
            myCounterCTUD[i].noDown = 0;
            switch(tempFlag){
              case 'V':
                  myCounterCTUD[i].noDownIndex = (dueFlashStorage.read(tempMulti+13))-48;
                break;

              case 'v':
                  myCounterCTUD[i].noDownIndex = (dueFlashStorage.read(tempMulti+13))-48+10;
                break;

              case 'U':
                  myCounterCTUD[i].noDownIndex = (dueFlashStorage.read(tempMulti+13))-48+20;
                break;

              case 'u':
                  myCounterCTUD[i].noDownIndex = (dueFlashStorage.read(tempMulti+13))-48+30;
                break;

              case 'A':
                  myCounterCTUD[i].noDownIndex = (dueFlashStorage.read(tempMulti+13))-48+40;
                   break;

                 case 'a':
                  myCounterCTUD[i].noDownIndex = (dueFlashStorage.read(tempMulti+13))-48+50;
                 break;

                 case 'c':
                  myCounterCTUD[i].noDownIndex = (dueFlashStorage.read(tempMulti+13))-48+60;
                break;
            }//END switch
          }


          tempFlag = dueFlashStorage.read(tempMulti+14);
          if(tempFlag == 's'){
            myCounterCTUD[i].countDownNull = true;
          }
          else{
            myCounterCTUD[i].countDownNull = false;
            if(tempFlag == 'O') {
              myCounterCTUD[i].countDown = true;
              myCounterCTUD[i].countDownIndex = (dueFlashStorage.read(tempMulti+15))-48;
            }
            else if(tempFlag == 'k'){
              myCounterCTUD[i].countDown = true;
              myCounterCTUD[i].countDownIndex = (dueFlashStorage.read(tempMulti+15))-48+10;
            }
            else {
              myCounterCTUD[i].countDown = false;
              switch(tempFlag){
                case 'V':
                    myCounterCTUD[i].countDownIndex = (dueFlashStorage.read(tempMulti+15))-48;
                  break;

                case 'v':
                    myCounterCTUD[i].countDownIndex = (dueFlashStorage.read(tempMulti+15))-48+10;
                  break;

                case 'U':
                    myCounterCTUD[i].countDownIndex = (dueFlashStorage.read(tempMulti+15))-48+20;
                  break;

                case 'u':
                    myCounterCTUD[i].countDownIndex = (dueFlashStorage.read(tempMulti+15))-48+30;
                  break;

               case 'A':
                  myCounterCTUD[i].countDownIndex = (dueFlashStorage.read(tempMulti+15))-48+40;
                   break;

                 case 'a':
                  myCounterCTUD[i].countDownIndex = (dueFlashStorage.read(tempMulti+15))-48+50;
                 break;

                 case 'c':
                  myCounterCTUD[i].countDownIndex = (dueFlashStorage.read(tempMulti+15))-48+60;
                break;
              }//END switch
            }
          }//END else

        }//END else if(tempCounterValue == 'b')



      }//END for Counters


      for(int i = 0; i<24; i++){                                                         //Loading Timers
        int tempMulti = REG_TIMER + (12*i);
        char tempCounterValue = dueFlashStorage.read(tempMulti);


        if(tempCounterValue == 't'){                                                                           //TON
          myTimer[i].typeTimer = 1;//TON

          int tempFlag = dueFlashStorage.read(tempMulti+1);
          if(tempFlag == 'I') {
            myTimer[i].no = 1;
            myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48;
          }
          else if(tempFlag == 'i'){
            myTimer[i].no = 1;
            myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+10;
          }
          else if(tempFlag == 'O'){
            myTimer[i].no = 2;
            myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48;
          }
          else if(tempFlag == 'k'){
            myTimer[i].no = 2;
            myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+10;
          }
          else {
            myTimer[i].no = 0;
            switch(tempFlag){
                case 'V':
                    myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48;
                  break;

                case 'v':
                    myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+10;
                  break;

                case 'U':
                    myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+20;
                  break;

                case 'u':
                    myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+30;
                  break;

                case 'A':
                  myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+40;
                   break;

                 case 'a':
                  myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+50;
                 break;

                 case 'c':
                  myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+60;
                break;
              }//END switch
          }//END else


          tempFlag = dueFlashStorage.read(tempMulti+3);
          if(tempFlag == 's'){
            myTimer[i].dnNull = true;
          }
          else{
            myTimer[i].dnNull = false;
            if(tempFlag == 'O') {
              myTimer[i].dn = true;
              myTimer[i].dnIndex = (dueFlashStorage.read(tempMulti+4))-48;
            }
            else if(tempFlag == 'k'){
              myTimer[i].dn = true;
              myTimer[i].dnIndex = (dueFlashStorage.read(tempMulti+4))-48+10;
            }
            else {
              myTimer[i].dn = false;
              switch(tempFlag){
                case 'V':
                    myTimer[i].dnIndex = (dueFlashStorage.read(tempMulti+4))-48;
                  break;

                case 'v':
                    myTimer[i].dnIndex = (dueFlashStorage.read(tempMulti+4))-48+10;
                  break;

                case 'U':
                    myTimer[i].dnIndex = (dueFlashStorage.read(tempMulti+4))-48+20;
                  break;

                case 'u':
                    myTimer[i].dnIndex = (dueFlashStorage.read(tempMulti+4))-48+30;
                  break;

                case 'A':
                  myTimer[i].dnIndex = (dueFlashStorage.read(tempMulti+4))-48+40;
                   break;

                 case 'a':
                  myTimer[i].dnIndex = (dueFlashStorage.read(tempMulti+4))-48+50;
                 break;

                 case 'c':
                  myTimer[i].dnIndex = (dueFlashStorage.read(tempMulti+4))-48+60;
                break;
              }//END switch
            }//END else
          }//END else


          tempFlag = dueFlashStorage.read(tempMulti+5);
          if(tempFlag == 's'){
            myTimer[i].ttNull = true;
          }
          else{
            myTimer[i].ttNull = false;

            if(tempFlag == 'O') {
              myTimer[i].tt = true;
              myTimer[i].ttIndex = (dueFlashStorage.read(tempMulti+6))-48;
            }
            else if(tempFlag == 'k'){
              myTimer[i].tt = true;
              myTimer[i].ttIndex = (dueFlashStorage.read(tempMulti+6))-48+10;
            }
            else {
              myTimer[i].tt = false;
              switch(tempFlag){
                case 'V':
                    myTimer[i].ttIndex = (dueFlashStorage.read(tempMulti+6))-48;
                  break;

                case 'v':
                    myTimer[i].ttIndex = (dueFlashStorage.read(tempMulti+6))-48+10;
                  break;

                case 'U':
                    myTimer[i].ttIndex = (dueFlashStorage.read(tempMulti+6))-48+20;
                  break;

                case 'u':
                    myTimer[i].ttIndex = (dueFlashStorage.read(tempMulti+6))-48+30;
                  break;

                  case 'A':
                  myTimer[i].ttIndex = (dueFlashStorage.read(tempMulti+6))-48+40;
                   break;

                 case 'a':
                  myTimer[i].ttIndex = (dueFlashStorage.read(tempMulti+6))-48+50;
                 break;

                 case 'c':
                  myTimer[i].ttIndex = (dueFlashStorage.read(tempMulti+6))-48+60;
                break;
              }//END switch
            }//END else
          }//END else

          myTimer[i].presetIndex = dueFlashStorage.read(tempMulti+8);
          myTimer[i].currentTimeIndex = dueFlashStorage.read(tempMulti+9);

           

          if(dueFlashStorage.read(tempMulti+7)=='0')
            myTimer[i].isPoint1 = false;

          else if(dueFlashStorage.read(tempMulti+7)=='1')
            myTimer[i].isPoint1 = true;


          myTimer[i].enableNo = false;
          myTimer[i].activeTimer = false;
        }



        else if(tempCounterValue == 'T'){                                                                      //TOF
          myTimer[i].typeTimer = 2;//TOF

          int tempFlag = dueFlashStorage.read(tempMulti+1);
          if(tempFlag == 'I') {
            myTimer[i].no = 1;
            myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48;
          }
          else if(tempFlag == 'i'){
            myTimer[i].no = 1;
            myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+10;
          }
          else if(tempFlag == 'O'){
            myTimer[i].no = 2;
            myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48;
          }
          else if(tempFlag == 'k'){
            myTimer[i].no = 2;
            myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+10;
          }
          else {
            myTimer[i].no = 0;
            switch(tempFlag){
                case 'V':
                    myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48;
                  break;

                case 'v':
                    myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+10;
                  break;

                case 'U':
                    myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+20;
                  break;

                case 'u':
                    myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+30;
                  break;

                 case 'A':
                  myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+40;
                   break;

                 case 'a':
                  myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+50;
                 break;

                 case 'c':
                  myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+60;
                break;
              }//END switch
          }//END else


          tempFlag = dueFlashStorage.read(tempMulti+3);
          if(tempFlag == 's'){
            myTimer[i].dnNull = true;
          }
          else{
            myTimer[i].dnNull = false;
            if(tempFlag == 'O') {
              myTimer[i].dn = true;
              myTimer[i].dnIndex = (dueFlashStorage.read(tempMulti+4))-48;
            }
            else if(tempFlag == 'k'){
              myTimer[i].dn = true;
              myTimer[i].dnIndex = (dueFlashStorage.read(tempMulti+4))-48+10;
            }
            else {
              myTimer[i].dn = false;
              switch(tempFlag){
                case 'V':
                    myTimer[i].dnIndex = (dueFlashStorage.read(tempMulti+4))-48;
                  break;

                case 'v':
                    myTimer[i].dnIndex = (dueFlashStorage.read(tempMulti+4))-48+10;
                  break;

                case 'U':
                    myTimer[i].dnIndex = (dueFlashStorage.read(tempMulti+4))-48+20;
                  break;

                case 'u':
                    myTimer[i].dnIndex = (dueFlashStorage.read(tempMulti+4))-48+30;
                  break;

                case 'A':
                  myTimer[i].dnIndex = (dueFlashStorage.read(tempMulti+4))-48+40;
                   break;

                 case 'a':
                  myTimer[i].dnIndex = (dueFlashStorage.read(tempMulti+4))-48+50;
                 break;

                 case 'c':
                  myTimer[i].dnIndex = (dueFlashStorage.read(tempMulti+4))-48+60;
                break;
              }//END switch
            }//END else
          }//END else


          tempFlag = dueFlashStorage.read(tempMulti+5);
          if(tempFlag == 's'){
            myTimer[i].ttNull = true;
          }
          else{
            myTimer[i].ttNull = false;

            if(tempFlag == 'O') {
              myTimer[i].tt = true;
              myTimer[i].ttIndex = (dueFlashStorage.read(tempMulti+6))-48;
            }
            else if(tempFlag == 'k'){
              myTimer[i].tt = true;
              myTimer[i].ttIndex = (dueFlashStorage.read(tempMulti+6))-48+10;
            }
            else {
              myTimer[i].tt = false;
              switch(tempFlag){
                case 'V':
                    myTimer[i].ttIndex = (dueFlashStorage.read(tempMulti+6))-48;
                  break;

                case 'v':
                    myTimer[i].ttIndex = (dueFlashStorage.read(tempMulti+6))-48+10;
                  break;

                case 'U':
                    myTimer[i].ttIndex = (dueFlashStorage.read(tempMulti+6))-48+20;
                  break;

                case 'u':
                    myTimer[i].ttIndex = (dueFlashStorage.read(tempMulti+6))-48+30;
                  break;

                 case 'A':
                  myTimer[i].ttIndex = (dueFlashStorage.read(tempMulti+6))-48+40;
                   break;

                 case 'a':
                  myTimer[i].ttIndex = (dueFlashStorage.read(tempMulti+6))-48+50;
                 break;

                 case 'c':
                  myTimer[i].ttIndex = (dueFlashStorage.read(tempMulti+6))-48+60;
                break;
              }//END switch
            }//END else
          }//END else


          myTimer[i].presetIndex = dueFlashStorage.read(tempMulti+8);
          myTimer[i].currentTimeIndex = dueFlashStorage.read(tempMulti+9);

          if(dueFlashStorage.read(tempMulti+7)=='0')
            myTimer[i].isPoint1 = false;

          else if(dueFlashStorage.read(tempMulti+7)=='1')
            myTimer[i].isPoint1 = true;

          myTimer[i].enableNo = false;
          myTimer[i].activeTimer = false;
        }


        else if(tempCounterValue == 'S'){                                                                      //RTO

          myTimer[i].typeTimer = 3;//RTO

          int tempFlag = dueFlashStorage.read(tempMulti+1);
          if(tempFlag == 'I') {
            myTimer[i].no = 1;
            myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48;
          }
          else if(tempFlag == 'i'){
            myTimer[i].no = 1;
            myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+10;
          }
          else if(tempFlag == 'O'){
            myTimer[i].no = 2;
            myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48;
          }
          else if(tempFlag == 'k'){
            myTimer[i].no = 2;
            myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+10;
          }
          else {
            myTimer[i].no = 0;
            switch(tempFlag){
                case 'V':
                    myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48;
                  break;

                case 'v':
                    myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+10;
                  break;

                case 'U':
                    myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+20;
                  break;

                case 'u':
                    myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+30;
                  break;

                 case 'A':
                  myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+40;
                   break;

                 case 'a':
                  myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+50;
                 break;

                 case 'c':
                  myTimer[i].noIndex = (dueFlashStorage.read(tempMulti+2))-48+60;
                break;
              }//END switch
          }//END else


          tempFlag = dueFlashStorage.read(tempMulti+3);
          if(tempFlag == 's'){
            myTimer[i].dnNull = true;
          }
          else{
            myTimer[i].dnNull = false;
            if(tempFlag == 'O') {
              myTimer[i].dn = true;
              myTimer[i].dnIndex = (dueFlashStorage.read(tempMulti+4))-48;
            }
            else if(tempFlag == 'k'){
              myTimer[i].dn = true;
              myTimer[i].dnIndex = (dueFlashStorage.read(tempMulti+4))-48+10;
            }
            else {
              myTimer[i].dn = false;
              switch(tempFlag){
                case 'V':
                    myTimer[i].dnIndex = (dueFlashStorage.read(tempMulti+4))-48;
                  break;

                case 'v':
                    myTimer[i].dnIndex = (dueFlashStorage.read(tempMulti+4))-48+10;
                  break;

                case 'U':
                    myTimer[i].dnIndex = (dueFlashStorage.read(tempMulti+4))-48+20;
                  break;

                case 'u':
                    myTimer[i].dnIndex = (dueFlashStorage.read(tempMulti+4))-48+30;
                  break;

                case 'A':
                  myTimer[i].dnIndex = (dueFlashStorage.read(tempMulti+4))-48+40;
                   break;

                 case 'a':
                  myTimer[i].dnIndex = (dueFlashStorage.read(tempMulti+4))-48+50;
                 break;

                 case 'c':
                  myTimer[i].dnIndex = (dueFlashStorage.read(tempMulti+4))-48+60;
                break;
              }//END switch
            }//END else
          }//END else


          tempFlag = dueFlashStorage.read(tempMulti+5);
          if(tempFlag == 's'){
            myTimer[i].ttNull = true;
          }
          else{
            myTimer[i].ttNull = false;

            if(tempFlag == 'O') {
              myTimer[i].tt = true;
              myTimer[i].ttIndex = (dueFlashStorage.read(tempMulti+6))-48;
            }
            else if(tempFlag == 'k'){
              myTimer[i].tt = true;
              myTimer[i].ttIndex = (dueFlashStorage.read(tempMulti+6))-48+10;
            }
            else {
              myTimer[i].tt = false;
              switch(tempFlag){
                case 'V':
                    myTimer[i].ttIndex = (dueFlashStorage.read(tempMulti+6))-48;
                  break;

                case 'v':
                    myTimer[i].ttIndex = (dueFlashStorage.read(tempMulti+6))-48+10;
                  break;

                case 'U':
                    myTimer[i].ttIndex = (dueFlashStorage.read(tempMulti+6))-48+20;
                  break;

                case 'u':
                    myTimer[i].ttIndex = (dueFlashStorage.read(tempMulti+6))-48+30;
                  break;

                case 'A':
                  myTimer[i].ttIndex = (dueFlashStorage.read(tempMulti+6))-48+40;
                   break;

                 case 'a':
                  myTimer[i].ttIndex = (dueFlashStorage.read(tempMulti+6))-48+50;
                 break;

                 case 'c':
                  myTimer[i].ttIndex = (dueFlashStorage.read(tempMulti+6))-48+60;
                break;
              }//END switch
            }//END else
          }//END else


          myTimer[i].presetIndex = dueFlashStorage.read(tempMulti+8);
          myTimer[i].currentTimeIndex = dueFlashStorage.read(tempMulti+9);

          if(dueFlashStorage.read(tempMulti+7)=='0')
            myTimer[i].isPoint1 = false;

          else if(dueFlashStorage.read(tempMulti+7)=='1')
            myTimer[i].isPoint1 = true;


          tempFlag = dueFlashStorage.read(tempMulti+10);
          if(tempFlag == 'I') {
            myTimer[i].reset = 1;
            myTimer[i].resetIndex = (dueFlashStorage.read(tempMulti+11))-48;
          }
          else if(tempFlag == 'i'){
            myTimer[i].reset = 1;
            myTimer[i].resetIndex = (dueFlashStorage.read(tempMulti+11))-48+10;
          }
          else if(tempFlag == 'O'){
            myTimer[i].reset = 2;
            myTimer[i].resetIndex = (dueFlashStorage.read(tempMulti+11))-48;
          }
          else if(tempFlag == 'k'){
            myTimer[i].reset = 2;
            myTimer[i].resetIndex = (dueFlashStorage.read(tempMulti+11))-48+10;
          }
          else {
            myTimer[i].reset = 0;
            switch(tempFlag){
                case 'V':
                    myTimer[i].resetIndex = (dueFlashStorage.read(tempMulti+11))-48;
                  break;

                case 'v':
                    myTimer[i].resetIndex = (dueFlashStorage.read(tempMulti+11))-48+10;
                  break;

                case 'U':
                    myTimer[i].resetIndex = (dueFlashStorage.read(tempMulti+11))-48+20;
                  break;

                case 'u':
                    myTimer[i].resetIndex = (dueFlashStorage.read(tempMulti+11))-48+30;
                  break;

                case 'A':
                  myTimer[i].resetIndex = (dueFlashStorage.read(tempMulti+11))-48+40;
                   break;

                 case 'a':
                  myTimer[i].resetIndex = (dueFlashStorage.read(tempMulti+11))-48+50;
                 break;

                 case 'c':
                  myTimer[i].resetIndex = (dueFlashStorage.read(tempMulti+11))-48+60;
                break;
              }//END switch
          }//END else

          myTimer[i].enableNo = false;
          myTimer[i].activeTimer = false;
          dintValues[myTimer[i].currentTimeIndex] = 0;

        }

      }//END FOR Timers

    }//END IF

    else{
      digitalWrite(13, LOW);
    }

    PC=0;//HERE
    n=0;
    X=false;
    Y=false;
    notFlag=false;
    fallingFlag=false;
    risingFlag=false;
    iFlag=false;
    iFlag1=false;
    oFlag=false;
    oFlag1=false;
    vFlag=false;
    vFlag1=false;
    vFlag2=false;
    vFlag3=false;
    vFlag4=false;
    vFlag5=false;
    vFlag6=false;
    latchFlag=false;
    unlatchFlag=false;
    negateFlag=false;
    coilFlag=false;
    timerFlag=false;
    counterFlag=false;
    counterCTUDFlag = false;

    inFunction = false;
                                                                //notFunction mode
  }

void reprogram(){

    if(Serial1.available() > 0){

      if(inReprogramTime){
        Serial1.readBytesUntil(42, bufferSerial, sizeof(bufferSerial));                            

        if(bufferSerial[0] == 'c'){
          swapLState();
          Serial1.println("a");
          delay(10);

          bool programing = true;
          bool updateMyTime = false;                                                                    
          int countPC= 0;
          int countCounter = 0;
          int countCounterCTU = 0;
          int countCounterCTUD = 0;
          int countTimer = 0;
          int counterInt = 0;
          int counterDint = 0;
          int counterReal = 0;
          int varSalto = 0;
          int32_t tempA = 0;
          int32_t tempB = 0;
          startTime=millis();

          while(programing){

          if(Serial1.available()>0){
            buffSize = Serial1.readBytesUntil(42,bufferSerial,sizeof(bufferSerial));     //42 *
            if(buffSize > 0){
              switch(bufferSerial[0]){
                case 'I': //I
                    for(int i = 1; i<buffSize; i++){
                      dueFlashStorage.write(REG_IN+i-1,bufferSerial[i]);
                    }
                  break;

                case 'O': //O
                    for(int i = 1; i<buffSize; i++){
                      dueFlashStorage.write(REG_OUT+i-1,bufferSerial[i]);
                    }
                  break;

                case 'V': //V
                    for(int i = 1; i<buffSize; i++){
                      dueFlashStorage.write(REG_VIRTUAL+i-1,bufferSerial[i]);
                    }
                  break;

                case 'w':
                    for(int i = 1; i<buffSize; i++){
                      dueFlashStorage.write(REG_TYPE_SIZE+i-1, bufferSerial[i]);
                    }
                  break;

                case 'N': //int
                
                    varSalto = (REG_INT + (2 * counterInt));

                    tempB = bufferSerial[3] << 14;
                    tempA = bufferSerial[2] << 7;
                    tempB = tempB | tempA | bufferSerial[1];

                    tempA = tempB & (0xFF);
                    dueFlashStorage.write(varSalto + 1, tempA);

                    tempA = tempB & (0xFF00);
                    tempA = tempA >> 8;

                    dueFlashStorage.write(varSalto, tempA);

                    counterInt++;
                  break;

                case 'L': //dint
                      varSalto = (REG_LONG+(4 * counterDint));

                      tempB = (((int32_t)bufferSerial[5]) << 28);
                      tempA = (((int32_t)bufferSerial[4]) << 21);
                      tempB = tempB | tempA;
                      tempA = (((int32_t)bufferSerial[3]) << 14);
                      tempB = tempB | tempA;
                      tempA = (((int32_t)bufferSerial[2]) << 7);
                      tempB = tempB | tempA | ((int32_t)bufferSerial[1]);

                      tempA = tempB & (0xFF);
                      dueFlashStorage.write(varSalto + 3, tempA);

                      tempA = tempB & (0xFF00);
                      tempA = tempA >> 8;
                      dueFlashStorage.write(varSalto + 2, tempA);

                      tempA = tempB & (0xFF0000);
                      tempA = tempA >> 16;
                      dueFlashStorage.write(varSalto + 1, tempA);

                      tempA = tempB & (0xFF000000);
                      tempA = tempA >> 24;
                      dueFlashStorage.write(varSalto, tempA);


                      counterDint++;
                  break;

                case 'F': //real
                      varSalto = (REG_FLOAT) + (4 * counterReal);

                      tempB = (((int32_t)bufferSerial[5]) << 28);
                      tempA = (((int32_t)bufferSerial[4]) << 21);
                      tempB = tempB | tempA;
                      tempA = (((int32_t)bufferSerial[3]) << 14);
                      tempB = tempB | tempA;
                      tempA = (((int32_t)bufferSerial[2]) << 7);
                      tempB = tempB | tempA | ((int32_t)bufferSerial[1]);

                      tempA = tempB & (0xFF);
                      dueFlashStorage.write(varSalto + 3, tempA);

                      tempA = tempB & (0xFF00);
                      tempA = tempA >> 8;
                      dueFlashStorage.write(varSalto + 2, tempA);

                      tempA = tempB & (0xFF0000);
                      tempA = tempA >> 16;
                      dueFlashStorage.write(varSalto + 1, tempA);

                      tempA = tempB & (0xFF000000);
                      tempA = tempA >> 24;
                      dueFlashStorage.write(varSalto, tempA);

                    counterReal++;
                  break;

                case '$':
                  for(int i = 0; i<buffSize; i++){
                    dueFlashStorage.write(countPC, bufferSerial[i]);
                    countPC++;
                  }
                break;

                case 'R':
                  for(int i = 0; i<buffSize; i++){
                    dueFlashStorage.write(countPC, bufferSerial[i]);
                    countPC++;
                  }
                  break;

                case 'B':
                  for(int i = 0; i<buffSize; i++){
                    int salto = (REG_COUNTER+i)+(16*countCounter);
                    dueFlashStorage.write(salto, bufferSerial[i]);
                  }
                  dueFlashStorage.write(countPC, 'B');
                  countPC++;
                  dueFlashStorage.write(countPC, (countCounter+48)); //countCounterCTU
                  countPC++;
                  countCounter++;
                  countCounterCTU++;
                  break;

                case 'b':
                  for(int i = 0; i<buffSize; i++){
                    int salto = (REG_COUNTER+i)+(16*countCounter);
                    dueFlashStorage.write(salto, bufferSerial[i]);
                  }
                  dueFlashStorage.write(countPC, 'b');
                  countPC++;
                  dueFlashStorage.write(countPC, (countCounter+48));  //countCounterCTUD
                  countPC++;
                  countCounter++;
                  countCounterCTUD++;
                  break;

                case 't':
                  for(int i = 0; i<buffSize; i++){
                    int salto = (REG_TIMER+i)+(12*countTimer);
                    dueFlashStorage.write(salto, bufferSerial[i]);
                  }
                  dueFlashStorage.write(countPC,'q');
                  countPC++;
                  dueFlashStorage.write(countPC, (countTimer+48));
                  countPC++;
                  countTimer++;
                  break;

                case 'T':
                  for(int i = 0; i<buffSize; i++){
                    int salto = (REG_TIMER+i)+(12*countTimer);
                    dueFlashStorage.write(salto, bufferSerial[i]);
                  }
                  dueFlashStorage.write(countPC,'q');
                  countPC++;
                  dueFlashStorage.write(countPC, (countTimer+48));
                  countPC++;
                  countTimer++;
                  break;

                case 'S': 
                  for(int i = 0; i<buffSize; i++){
                    int salto = (REG_TIMER+i)+(12*countTimer);
                    dueFlashStorage.write(salto, bufferSerial[i]);
                  }
                  dueFlashStorage.write(countPC,'q');
                  countPC++;
                  dueFlashStorage.write(countPC, (countTimer+48));
                  countPC++;
                  countTimer++;
                  break;

                case'x':
                  Serial1.flush();
                  while(Serial1.read() != -1){

                  }

                  Serial1.println('y');      
                 
                  break;

                case'X':
                  Serial1.flush();
                  while(Serial1.read() != -1){

                  }
                  Serial1.println('h');
       
                  break;

                case 'r':
                  for(int i = 1; i<buffSize;i++){
                    dueFlashStorage.write(countPC, bufferSerial[i]);
                    countPC++;
                  }
                  break;

                case 'Y':  
                    tempB = bufferSerial[3] << 14;
                    tempA = bufferSerial[2] << 7;
                    tempB = tempB | tempA | bufferSerial[1];

                    tempA = tempB & (0xFF);
                    timeData[0] = tempA;
                  break;

                case 'M':
                    tempB = bufferSerial[3] << 14;
                    tempA = bufferSerial[2] << 7;
                    tempB = tempB | tempA | bufferSerial[1];

                    tempA = tempB & (0xFF);
                    timeData[1] = tempA;
                  break;

                case 'D':
                    tempB = bufferSerial[3] << 14;
                    tempA = bufferSerial[2] << 7;
                    tempB = tempB | tempA | bufferSerial[1];

                    tempA = tempB & (0xFF);
                    timeData[3] = tempA;
                  break;

                case 'H':
                    tempB = bufferSerial[3] << 14;
                    tempA = bufferSerial[2] << 7;
                    tempB = tempB | tempA | bufferSerial[1];

                    tempA = tempB & (0xFF);
                    timeData[4] = tempA;
                  break;

                case 'P':
                    tempB = bufferSerial[3] << 14;
                    tempA = bufferSerial[2] << 7;
                    tempB = tempB | tempA | bufferSerial[1];

                    tempA = tempB & (0xFF);
                    timeData[5] = tempA;
                  break;

                case 'k':
                    tempB = bufferSerial[3] << 14;
                    tempA = bufferSerial[2] << 7;
                    tempB = tempB | tempA | bufferSerial[1];

                    tempA = tempB & (0xFF);
                    timeData[6] = tempA;

                    updateMyTime = true;
                  break;
                  
                case 'd':
                    programing = false;
                    digitalWrite(13,HIGH);
                    dueFlashStorage.write(countPC, 'z');
                    dueFlashStorage.write(REG_STATUS,'2');

                    Serial1.flush();
                    while(Serial1.read() != -1){

                    }
                    cleanOldCounters(countCounter);
                    cleanOldTimers(countTimer);

                    startReprogramTime =  millis();
                    startBlinkReproTime = millis();

                  break;
              }//END switch bufferSerial[0]
              delay(20);
              swapLState();
            }//END buffSize > 0
          }//END if

        }//END while(programing)

        if(updateMyTime)                                                                        
          updateAllTime();
        
        initProgram();
        digitalWrite(13, HIGH);

        } //END if(bufferSerial[0] == 'c')
        
        else if(bufferSerial[0]=='C'){                                                        
          digitalWrite(13, LOW);
          Serial1.println("t");
          delay(10);

          int32_t tempA = 0;
          int32_t tempB = 0;
          bool updatingTime = true;


          while(updatingTime){
            if(Serial1.available()>0){
              
              buffSize = Serial1.readBytesUntil(42,bufferSerial,sizeof(bufferSerial));    

              if(buffSize > 0){
                switch(bufferSerial[0]){

                  case'x':
                  Serial1.flush();
                  while(Serial1.read() != -1){

                  }
                  Serial1.println("p");
                  
                  break;

                  case 'Y':                                                                               
                    tempB = bufferSerial[3] << 14;
                    tempA = bufferSerial[2] << 7;
                    tempB = tempB | tempA | bufferSerial[1];

                    tempA = tempB & (0xFF);
                    timeData[0] = tempA;
                   
                  break;

                case 'M':
                    tempB = bufferSerial[3] << 14;
                    tempA = bufferSerial[2] << 7;
                    tempB = tempB | tempA | bufferSerial[1];

                    tempA = tempB & (0xFF);
                    timeData[1] = tempA;
                   
                  break;

                case 'D':
                    tempB = bufferSerial[3] << 14;
                    tempA = bufferSerial[2] << 7;
                    tempB = tempB | tempA | bufferSerial[1];

                    tempA = tempB & (0xFF);
                    timeData[3] = tempA;
             
                  break;

                case 'H':
                    tempB = bufferSerial[3] << 14;
                    tempA = bufferSerial[2] << 7;
                    tempB = tempB | tempA | bufferSerial[1];

                    tempA = tempB & (0xFF);
                    timeData[4] = tempA;
          
                  break;

                case 'P':
                    tempB = bufferSerial[3] << 14;
                    tempA = bufferSerial[2] << 7;
                    tempB = tempB | tempA | bufferSerial[1];

                    tempA = tempB & (0xFF);
                    timeData[5] = tempA;
             
                  break;

                case 'k':
                    tempB = bufferSerial[3] << 14;
                    tempA = bufferSerial[2] << 7;
                    tempB = tempB | tempA | bufferSerial[1];

                    tempA = tempB & (0xFF);
                    timeData[6] = tempA;
                

                  break;

                case 'd':
                 
                    updatingTime = false;

                    Serial1.flush();
                    while(Serial1.read() != -1){

                    }

                  break;

                }//END switch
              }//END if
            }//END if

          }//END while

          updateAllTime();

          for(int i=0; i<3; i++){
            digitalWrite(13,HIGH);
            delay(300);
            digitalWrite(13,LOW);
            delay(300);
          }

          digitalWrite(13, HIGH);  
         }//END else if(bufferSerial[0] == 'C')


      } //END if(inReprogramTime)

      else{                                       //HMI mode

        int buffSize = 0;

        buffSize = Serial1.readBytesUntil(98,bufferSerial,sizeof(bufferSerial));

        if(buffSize > 0){

          if(!configStream){

              if((bufferSerial[0]=='c')){
                sendDataConfig = false;
                Serial1.write(100);
                Serial1.write('s');

                int tempSizeI = 0;
                int tempSizeO = 0;

                for(int i=0; i<16; i++){
                  if(iMask[i] == true)
                    tempSizeI++;
                  if(oMask[i] == true)
                    tempSizeO++;
                }

                fragmentData(tempSizeI, false, false);
                Serial1.write(md);
                Serial1.write(ld);

                Serial1.write('i');
                if(tempSizeI > 0){
                  for(int i=0; i<16; i++){
                    if(iMask[i] == true)
                      Serial1.write(i);
                  }
                }

                fragmentData(tempSizeO, false, false);
                Serial1.write(md);
                Serial1.write(ld);

                Serial1.write('o');

                if(tempSizeO > 0){
                  for(int i=0; i<16; i++){
                    if(oMask[i] == true)
                      Serial1.write(i);
                  }
                }

                tempSizeI = 0;
                for(int i=0; i<64; i++){
                  if(vMask[i] == true)
                    tempSizeI++;
                }

                fragmentData(tempSizeI, false, false);
                Serial1.write(md);
                Serial1.write(ld);

                Serial1.write('v');

                if(tempSizeI > 0){
                  for(int i=0; i<64; i++){
                    if(vMask[i] == true){
                        fragmentData(i, false, false);
                        Serial1.write(md);
                        Serial1.write(ld);
                    }
                  }
                }

                fragmentData((uint16_t)dueFlashStorage.read(REG_TYPE_SIZE), false, false);
                Serial1.write(md);
                Serial1.write(ld);

                fragmentData((uint16_t)dueFlashStorage.read(REG_TYPE_SIZE+1), false, false);
                Serial1.write(md);
                Serial1.write(ld);

                fragmentData((uint16_t)dueFlashStorage.read(REG_TYPE_SIZE+2), false, false);
                Serial1.write(md);
                Serial1.write(ld);

                Serial1.write(98);
                Serial1.flush();

                //fragmentData()
              }

              else if(bufferSerial[0]=='f'){
                //Serial1.flush();
                configStream = true;
                sendStartSend = true;
              }


          }//if(!configStream)

          else{
            if((bufferSerial[0]=='w')){
                inProcessLock = true;
                bool oOk = false, vOk = false, eOk = false, dOk = false, floatOk=false;
                int cursor = 1;
                int newCursor = 0;
                int dataIndex = 0;

                int tempValue = joinInt8(bufferSerial[cursor], bufferSerial[cursor + 1]);
                int tempType = bufferSerial[cursor + 2];
                cursor += 3;


                if(tempType == 'o'){
                  if(tempValue > 0){
                    for(int i=0; i<tempValue; i++){
                      dataIndex = bufferSerial[cursor+(i*2)];
                      if(bufferSerial[cursor+(i*2)+1] == 1){
                        oState[dataIndex] = true;
                      }
                      else{
                        oState[dataIndex] = false;
                      }

                      newCursor = cursor+(i*2)+2;

                    }
                    cursor = newCursor;
                  }
                  oOk = true;
                }

                tempValue = joinInt8(bufferSerial[cursor], bufferSerial[cursor + 1]);
                tempType = bufferSerial[cursor + 2];
                cursor +=3;

                if(tempType == 'v'){
                  if(tempValue > 0){
                    for(int i=0; i<tempValue; i++){
                      dataIndex = bufferSerial[cursor+(i*2)];

                      if(bufferSerial[cursor+(i*2)+1] == 1){
                        vState[dataIndex] = true;
                      }
                      else{
                        vState[dataIndex] = false;
                      }

                      newCursor = cursor+(i*2)+2;

                    }
                    cursor = newCursor;
                  }
                  vOk = true;
                }


                tempValue = joinInt8(bufferSerial[cursor], bufferSerial[cursor + 1]);
                tempType = bufferSerial[cursor + 2];
                cursor +=3;

                if(tempType == 'e'){
                  if(tempValue > 0){
                    for(int i=0; i<tempValue; i++){
                      dataIndex = bufferSerial[cursor+(i*4)];

                      int tempValue2 = joinInt16(bufferSerial[cursor+(i*4)+1], bufferSerial[cursor+(i*4)+2], bufferSerial[cursor+(i*4)+3]);
                      intValues[dataIndex] = tempValue2;

                      newCursor = cursor+(i*4)+4;

                    }
                    cursor = newCursor;
                  }
                  eOk = true;
                }

                tempValue = joinInt8(bufferSerial[cursor], bufferSerial[cursor + 1]);
                tempType = bufferSerial[cursor + 2];
                cursor +=3;

                if(tempType == 'd'){
                  if(tempValue > 0){
                    for(int i=0; i<tempValue; i++){
                      dataIndex = bufferSerial[cursor+(i*7)];

                      int tempValue2 = joinInt32(bufferSerial[cursor+(i*7)+1], bufferSerial[cursor+(i*7)+2], bufferSerial[cursor+(i*7)+3], bufferSerial[cursor+(i*7)+4],bufferSerial[cursor+(i*7)+5], bufferSerial[cursor+(i*7)+6]);
                      dintValues[dataIndex] = tempValue2;

                      newCursor = cursor+(i*7)+7;

                    }
                    cursor = newCursor;
                  }
                  dOk = true;
                }

                tempValue = joinInt8(bufferSerial[cursor], bufferSerial[cursor + 1]);
                tempType = bufferSerial[cursor + 2];
                cursor +=3;

                if(tempType == 'f'){
                  if(tempValue > 0){
                    for(int i=0; i<tempValue; i++){
                      dataIndex = bufferSerial[cursor+(i*7)];

                      uint32_t tempUint32 = joinInt32(bufferSerial[cursor+(i*7)+1], bufferSerial[cursor+(i*7)+2], bufferSerial[cursor+(i*7)+3], bufferSerial[cursor+(i*7)+4],bufferSerial[cursor+(i*7)+5], bufferSerial[cursor+(i*7)+6]);
                      float tempFloat = joinFloat(tempUint32);
                      realValues[dataIndex] = tempFloat;

                      newCursor = cursor+(i*7)+7;

                    }
                    cursor = newCursor;
                  }
                  floatOk = true;
                }

                if(oOk && vOk && eOk && dOk && floatOk){
                    requestSecondPart = true;
                }
                else{
                  inProcessLock = false;
                }

            }//END if((bufferSerial[0]=='w'))

            else if(bufferSerial[0] == 'y'){
                inProcessLock = true;
                bool iOk = false, oOk = false, vOk = false, eOk = false, dOk = false, floatOk = false;

                int cursor = 1;
                int newCursor = 0;
                int dataIndex = 0;

                int tempValue = joinInt8(bufferSerial[cursor], bufferSerial[cursor + 1]);
                int tempType = bufferSerial[cursor + 2];
                cursor += 3;

                iSize = tempValue;

                if(tempType == 'i'){
                  if(tempValue > 0){
                    for(int i=0; i<tempValue; i++){
                      iIndex[i] = bufferSerial[cursor+i];

                      newCursor = cursor+i+1;
                    }
                    cursor = newCursor;
                  }
                  iOk = true;
                }

                tempValue = joinInt8(bufferSerial[cursor], bufferSerial[cursor + 1]);
                tempType = bufferSerial[cursor + 2];
                cursor +=3;

                oSize = tempValue;

                if(tempType == 'o'){
                  if(tempValue > 0){
                    for(int i=0; i<tempValue; i++){
                      dataIndex = bufferSerial[cursor+i];
                      oIndex[i] = dataIndex;

                      newCursor = cursor + i + 1;
                    }
                    cursor = newCursor;
                  }
                  oOk = true;
                }


                tempValue = joinInt8(bufferSerial[cursor], bufferSerial[cursor + 1]);
                tempType = bufferSerial[cursor + 2];
                cursor +=3;

                vSize = tempValue;

                if(tempType == 'v'){
                  if(tempValue > 0){
                    for(int i=0; i<tempValue; i++){
                      dataIndex = bufferSerial[cursor+i];
                      vIndex[i] = dataIndex;

                      newCursor = cursor+i+1;

                    }
                    cursor = newCursor;
                  }
                  vOk = true;
                }


                tempValue = joinInt8(bufferSerial[cursor], bufferSerial[cursor + 1]);
                tempType = bufferSerial[cursor + 2];
                cursor +=3;

                eSize = tempValue;

                if(tempType == 'e'){
                  if(tempValue > 0){
                    for(int i=0; i<tempValue; i++){
                      dataIndex = bufferSerial[cursor+(i)];
                      intIndex[i] = dataIndex;

                      newCursor = cursor+i+1;

                    }
                    cursor = newCursor;
                  }
                  eOk = true;
                }

                tempValue = joinInt8(bufferSerial[cursor], bufferSerial[cursor + 1]);
                tempType = bufferSerial[cursor + 2];
                cursor +=3;

                dSize = tempValue;

                if(tempType == 'd'){
                  if(tempValue > 0){
                    for(int i=0; i<tempValue; i++){
                      dataIndex = bufferSerial[cursor+i];
                      dintIndex[i] = dataIndex;

                      newCursor = cursor+i+1;

                    }
                    cursor = newCursor;
                  }
                  dOk = true;
                }

                tempValue = joinInt8(bufferSerial[cursor], bufferSerial[cursor + 1]);
                tempType = bufferSerial[cursor + 2];
                cursor +=3;

                floatSize = tempValue;

                if(tempType == 'f'){
                  if(tempValue > 0){
                    for(int i=0; i<tempValue; i++){
                      dataIndex = bufferSerial[cursor+i];
                      realIndex[i] = dataIndex;

                      newCursor = cursor+i+1;

                    }
                    cursor = newCursor;
                  }
                  floatOk = true;
                }

               
                if(iOk && oOk && vOk && eOk && dOk && floatOk){
                    Serial1.write('z');
                    Serial1.write('i');

                    if(iSize > 0){
                      for(int i=0; i<iSize; i++){
                        dataIndex = iIndex[i];
                        if(iState[dataIndex])
                          Serial1.write(2);
                        else
                          Serial1.write(1);
                      }//END for
                    }//END if(iSize > 0)

                    Serial1.write('o');

                    if(oSize > 0){
                      for(int i=0; i<oSize; i++){
                        dataIndex = oIndex[i];
                        if(oState[dataIndex])
                          Serial1.write(2);
                        else
                          Serial1.write(1);
                      }//END for
                    }//END if

                    Serial1.write('v');

                    if(vSize > 0){
                      for(int i=0; i<vSize; i++){
                        dataIndex = vIndex[i];
                        if(vState[dataIndex])
                          Serial1.write(2);
                        else
                          Serial1.write(1);
                      }//END for
                    }//END if

                    Serial1.write('e');

                    if(eSize > 0){
                      for(int i=0; i<eSize; i++){
                        dataIndex = intIndex[i];

                        int dataTemp = intValues[dataIndex];
                        fragmentData(dataTemp);
                        Serial1.write(hd);
                        Serial1.write(md);
                        Serial1.write(ld);
                      }//END for
                    }//END if

                    Serial1.write('d');

                    if(dSize > 0){
                      for(int i=0; i<dSize; i++){
                        dataIndex = dintIndex[i];

                        int dataTemp = dintValues[dataIndex];
                        fragmentData(dataTemp, (uint8_t)4);
                        Serial1.write(hd32);
                        Serial1.write(md32);
                        Serial1.write(ld32);
                        Serial1.write(hd);
                        Serial1.write(md);
                        Serial1.write(ld);
                      }//END for
                    }//END if

                    Serial1.write('f');

                    if(floatSize > 0){
                      for(int i=0; i<floatSize; i++){
                        dataIndex = realIndex[i];

                        floatValue tempFloat;
                      tempFloat.val= realValues[dataIndex];

                      uint32_t tempUint32 = (uint32_t)tempFloat.bytes[0];
                      tempUint32 |= ((uint32_t)tempFloat.bytes[1] << 8);
                      tempUint32 |= ((uint32_t)tempFloat.bytes[2] << 16);
                      tempUint32 |= ((uint32_t)tempFloat.bytes[3] << 24);

                      fragmentData(tempUint32, (uint8_t)4);
                      Serial1.write(hd32);
                      Serial1.write(md32);
                      Serial1.write(ld32);
                      Serial1.write(hd);
                      Serial1.write(md);
                      Serial1.write(ld);

                      }//END for
                    }//END if

                    Serial1.write('b');
                }

                delay(150);

                alreadySendIt = false;
                sendStartSend = true;
                inProcessLock = false;

            }//END else if(bufferSerial[0] == 'y')


          }//else of if(!configStream)


        }//END if(buffSize > 0)
        
        
      }//HMI mode


    }//END if(Serial1.available())


    if(runReprogramTime){
      
      unsigned long currentTimeTemp = millis(); 
      currentTimeTemp -= startReprogramTime;


      if(currentTimeTemp > targetReprogramTime){
        inReprogramTime = false;
        runReprogramTime = false;
        digitalWrite(13, HIGH);
      }

      currentTimeTemp = millis();
      currentTimeTemp -= startBlinkReproTime;


      if(currentTimeTemp > blinkTime){
        swapLState();
        startBlinkReproTime = millis();
      }
     
    }

    if(sendStartSend){
      
        Serial1.write('w');
        Serial1.write('b');

        sendStartSend = false;
        alreadySendIt = true;
        inSecondState = false;
        inProcessLock = false;
        startTimeSend = millis();

    }//END if(sendStartSend)


    if(requestSecondPart){
      
      Serial1.write('y');
      Serial1.write('b');
      requestSecondPart = false;
      inSecondState = true;

      startTimeSend = millis();
      inProcessLock = false;
    }//END if(requestSecondPart)


    if(alreadySendIt){
      if(!inProcessLock){
        unsigned long tempCurrent = millis();
        tempCurrent -= startTimeSend;
        if(tempCurrent > targetTimeSend){
          
          Serial1.write('w');
          Serial1.write('b');

          sendStartSend = false;
          alreadySendIt = true;
          inSecondState = false;
          inProcessLock = false;
          startTimeSend = millis();

        }//END if(tempCurrent > targetTimeSend)
      }//END if(!inProcessLock)
    }//END if(alreadySendIt)


}//END reprogram


void cleanOldCounters(int tempCountCounter){
  if(tempCountCounter < 25){
    for(int i = tempCountCounter; i<25; i++){
      int salto = (REG_COUNTER)+(16*i);
      dueFlashStorage.write(salto, '~');//

    }
  }
}//END cleanOldCounters


void cleanOldTimers(int tempCountTimer){
  if(tempCountTimer< 25){
    for(int i = tempCountTimer; i<25; i++){
      int salto = (REG_TIMER)+(12*i);
      dueFlashStorage.write(salto, '~');//~

    }
  }
}//END cleanOldCounters


bool setDnCounter(int posi, bool dnState){
  bool tempReturn = dnState;
  if(!dnState){
    if(dintValues[myCounter[posi].countIndex] >= dintValues[myCounter[posi].presetIndex]){
      tempReturn = true;
    }//END if(myCounter[posi].count >= myCounter[posi].preset)
  }//END if
  else{
    if(myCounter[posi].resetNo){
      tempReturn = false;
    }
  }
  return tempReturn;
}

bool setDnCounterCTUD(int posi, bool dnState){
  bool tempReturn = dnState;
  if(!dnState){
    if(dintValues[myCounterCTUD[posi].countIndex] >= dintValues[myCounterCTUD[posi].presetIndex]){
      tempReturn = true;
    }//END if(myCounterCTUD[posi].count >= myCounterCTUD[posi].preset)
  }//END if
  else{
    if(myCounterCTUD[posi].resetNo){
      tempReturn = false;
    }

    if(dintValues[myCounterCTUD[posi].countIndex] < dintValues[myCounterCTUD[posi].presetIndex]){
      tempReturn = false;
    }//END if(myCounterCTUD[posi].count < myCounterCTUD[posi].preset)
  }
  return tempReturn;
}


void blinking(){
  if(digitalRead(13)){
      digitalWrite(13, LOW);
    }

    else{
      digitalWrite(13,HIGH);
    }
}

int readAnalogPin(int index){
  analogReadResolution(12);
  return analogRead(iPins[index]);
}

void writeServoPin(int index, int value){

  if(!servoOutput[index]){
    servoOutput[index] = true;

    switch(index){
      case 0:
        myServo0.attach(oPins[index]);
      break;
      case 1:
        myServo1.attach(oPins[index]);
      break;
      case 2:
        myServo2.attach(oPins[index]);
      break;
      case 3:
        myServo3.attach(oPins[index]);
      break;
      case 4:
        myServo4.attach(oPins[index]);
      break;
      case 5:
        myServo5.attach(oPins[index]);
      break;
      case 6:
        myServo6.attach(oPins[index]);
      break;
      case 7:
        myServo7.attach(oPins[index]);
      break;
  
    }

  }

  if(value > 180)
    value = 180;
  if(value < 0)
    value = 0;

    switch(index){
      case 0:
        myServo0.write(value);
      break;
      case 1:
        myServo1.write(value);
      break;
      case 2:
        myServo2.write(value);
      break;
      case 3:
        myServo3.write(value);
      break;
      case 4:
        myServo4.write(value);
      break;
      case 5:
        myServo5.write(value);
      break;
      case 6:
        myServo6.write(value);
      break;
      case 7:
        myServo7.write(value);
      break;

    }
    delay(15);

    if(value == 0)
      value = 1;

    analogOutputs[index] = (int16_t)value;
}

void writeAnalogPin(int index, int value){

  if(value > 4095)
    value = 4095 ;
  if(value < 0)
    value = 0;

  analogWriteResolution(12);
  analogWrite(oPins[index], value);
  analogOutputs[index] = (int16_t)value;
  
  delay(20);
}

int readAnalogOutput(int index){
  return (int)analogOutputs[index];
}

void equ(){                                                                     //Equal
  int dnIndex = 0;
  float tempSource = 0;
  float tempSourceB = 0;
  bool tagIsEnable = false;

  PC++;
  dnIndex = PC;
  PC+=2;
  tempSource = getSource();
  PC++;

  //Serial.print("SourceA: ");
  //Serial.println(tempSource);

  tagIsEnable = isTagEnable(PC+3);

  if(tagIsEnable){
    PC+=3;
    tempSourceB = getSource();
  }
  else{
    tempSourceB = (float)getJoinValue();
    PC+=4;
  }

  //Serial.print("SourceB: ");
  //Serial.println(tempSourceB);

  if(tempSource == tempSourceB)
    setDnValue(true, dnIndex);
  else
    setDnValue(false, dnIndex);
}

void neq(){                                                                     //Not equal
  int dnIndex = 0;
  float tempSource = 0;
  float tempSourceB = 0;
  bool tagIsEnable = false;

  PC++;
  dnIndex = PC;
  PC+=2;
  tempSource = getSource();
  PC++;

  tagIsEnable = isTagEnable(PC+3);

  if(tagIsEnable){
    PC+=3;
    tempSourceB = getSource();
  }
  else{
    tempSourceB = (float)getJoinValue();
    PC+=4;
  }


  if(tempSource != tempSourceB)
    setDnValue(true, dnIndex);
  else
    setDnValue(false, dnIndex);
}

void les(){                                                                     //Less than
  int dnIndex = 0;
  float tempSource = 0;
  float tempSourceB = 0;
  bool tagIsEnable = false;

  PC++;
  dnIndex = PC;
  PC+=2;
  tempSource = getSource();
  PC++;

  tagIsEnable = isTagEnable(PC+3);

  if(tagIsEnable){
    PC+=3;
    tempSourceB = getSource();
  }
  else{
    tempSourceB = (float)getJoinValue();
    PC+=4;
  }


  if(tempSource < tempSourceB)
    setDnValue(true, dnIndex);
  else
    setDnValue(false, dnIndex);

  //Serial.println("les");
}

void leq(){                                                                     //Less and equal
  int dnIndex = 0;
  float tempSource = 0;
  float tempSourceB = 0;
  bool tagIsEnable = false;

  PC++;
  dnIndex = PC;
  PC+=2;
  tempSource = getSource();
  PC++;

  tagIsEnable = isTagEnable(PC+3);

  if(tagIsEnable){
    PC+=3;
    tempSourceB = getSource();
  }
  else{
    tempSourceB = (float)getJoinValue();
    PC+=4;
  }


  if(tempSource <= tempSourceB)
    setDnValue(true, dnIndex);
  else
    setDnValue(false, dnIndex);
}

void grt(){                                                                     //Greater than
  int dnIndex = 0;
  float tempSource = 0;
  float tempSourceB = 0;
  bool tagIsEnable = false;

  PC++;
  dnIndex = PC;
  PC+=2;
  tempSource = getSource();
  PC++;

  tagIsEnable = isTagEnable(PC+3);

  if(tagIsEnable){
    PC+=3;
    tempSourceB = getSource();
  }
  else{
    tempSourceB = (float)getJoinValue();
    PC+=4;
  }


  if(tempSource > tempSourceB)
    setDnValue(true, dnIndex);
  else
    setDnValue(false, dnIndex);

  //Serial.println("GRT");
}

void geq(){                                                                     //Greater and equal
  int dnIndex = 0;
  float tempSource = 0;
  float tempSourceB = 0;
  bool tagIsEnable = false;

  PC++;
  dnIndex = PC;
  PC+=2;
  tempSource = getSource();
  PC++;

  tagIsEnable = isTagEnable(PC+3);

  if(tagIsEnable){
    PC+=3;
    tempSourceB = getSource();
  }
  else{
    tempSourceB = (float)getJoinValue();
    PC+=4;
  }


  if(tempSource >= tempSourceB)
    setDnValue(true, dnIndex);
  else
    setDnValue(false, dnIndex);
}


void mov(){                                                                     //Move
    bool noState = false;

    PC++;

    if(isTagEnable(PC))
      noState = getNoValue();
    else{
      noState = true;
      PC++;
    }

    if(noState){
        float tempSource = 0;

        PC++;
        if(isTagEnable(PC)){
          tempSource = getSource();
          PC+=4;
        }
        else{
          PC+=2;
          tempSource = (float)getJoinValue();
          PC+=3;
        }

        //Serial.println("Mov: ");
        //Serial.println(tempSource);
        //Serial.println("");
        setDest(tempSource);
    }
    else
      PC+=7;

}

void lim(){                                                                     //Limit
  int dnIndex = 0;
  float tempTest = 0;
  float tempSourceMax = 0;
  float tempSourceMin = 0;
  bool tagIsEnable = false;

  PC++;                                                                         //byte 2
  dnIndex = PC;
  PC+=2;                                                                        //byte 4
  tempTest = getSource();
  PC++;                                                                         //byte 6
  tagIsEnable = isTagEnable(PC+3);

  if(tagIsEnable){
    PC+=3;                                                                      //byte 9
    tempSourceMin = getSource();

  }
  else{
    tempSourceMin = (float)getJoinValue();
    PC+=4;                                                                      //byte 10
  }
  PC++;                                                                         //byte 11

  tagIsEnable = isTagEnable(PC+3);

  if(tagIsEnable){
    PC+=3;                                                                      //byte 14
    tempSourceMax = getSource();
  }
  else{
    tempSourceMax = (float)getJoinValue();
    PC+=4;
  }


  if((tempTest >= tempSourceMin) && (tempTest <= tempSourceMax))
    setDnValue(true, dnIndex);
  else
    setDnValue(false, dnIndex);

  //Serial.println("lim");

}

void meq(){                                                                     //Mask equal
  int dnIndex = 0;
  int32_t tempSource = 0;
  int32_t tempMask = 0;
  int32_t tempCompare = 0;
  bool tagIsEnable = false;

  PC++;                                                                         //byte 2
  dnIndex = PC;
  PC+=2;                                                                        //byte 4
  tempSource = getSource();
  PC++;                                                                         //byte 6
  tagIsEnable = isTagEnable(PC+3);

  if(tagIsEnable){
    PC+=3;                                                                      //byte 9
    tempMask = getSource();

  }
  else{
    tempMask = (float)getJoinValue();
    PC+=4;                                                                      //byte 10
  }
  PC++;                                                                         //byte 11

  tagIsEnable = isTagEnable(PC+3);

  if(tagIsEnable){
    PC+=3;                                                                      //byte 14
    tempCompare = getSource();
  }
  else{
    tempCompare = (float)getJoinValue();
    PC+=4;
  }

tempSource = tempSource & tempMask;

  if(tempSource == tempCompare)
    setDnValue(true, dnIndex);
  else
    setDnValue(false, dnIndex);
}//END meq()

void andOp(){                                                                   //And
  bool noState = false;

  PC++;                                                                         //byte 2

  if(isTagEnable(PC))
    noState = getNoValue();
  else{
    noState = true;
    PC++;                                                                       //byte 3
  }


  if(noState){
      int32_t tempSourceA = 0;
      int32_t tempSourceB = 0;

      PC++;                                                                     //byte 4
      tempSourceA = (int32_t)getSource();
      PC++;                                                                     //byte 6

      if(isTagEnable(PC+3)){                                                    //(byte 9)
        PC+=3;                                                                  //byte 9
        tempSourceB = (int32_t)getSource();
        PC++;                                                                   //byte 11
      }
      else{
        tempSourceB = (int32_t)getJoinValue();
        PC+=5;                                                                  //byte 11
      }

      tempSourceA = tempSourceA & tempSourceB;

      setDest(tempSourceA);
  }
  else
    PC+=9;                                                                      //byte 12

}//END andOp()

void orOp(){                                                                    //Or
  bool noState = false;

  PC++;                                                                         //byte 2

  if(isTagEnable(PC))
    noState = getNoValue();
  else{
    noState = true;
    PC++;                                                                       //byte 3
  }


  if(noState){
      int32_t tempSourceA = 0;
      int32_t tempSourceB = 0;

      PC++;                                                                     //byte 4
      tempSourceA = (int32_t)getSource();
      PC++;                                                                     //byte 6

      if(isTagEnable(PC+3)){                                                    //(byte 9)
        PC+=3;                                                                  //byte 9
        tempSourceB = (int32_t)getSource();
        PC++;                                                                   //byte 11
      }
      else{
        tempSourceB = (int32_t)getJoinValue();
        PC+=5;                                                                  //byte 11
      }

      tempSourceA = tempSourceA | tempSourceB;

      setDest(tempSourceA);
  }
  else
    PC+=9;
}//END orOp()

void notOp(){                                                                   //Not
  bool noState = false;

  PC++;                                                                         //byte 2

  if(isTagEnable(PC))
    noState = getNoValue();                                                     //(byte 3)
  else{
    noState = true;
    PC++;                                                                       //byte 3
  }

  if(noState){
      int32_t tempSource32 = 0;
      int16_t tempSource16 = 0;
      float tempSource = 0;

      PC++;                                                                     //byte 4

      if(isSourceInt()){
        tempSource16 = (int16_t)getSource();
        tempSource16 = ~(tempSource16);
        tempSource = (float)tempSource16;
      }
      else{
        tempSource32 = (int32_t)getSource();
        tempSource32 = ~(tempSource32);
        tempSource = (float)tempSource32;
      }


      PC++;                                                                     //byte 6

      setDest(tempSource);
  }
  else
    PC+=4;                                                                      //byte 7
}//END notOp()

void xorOp(){                                                                   //Xor
  bool noState = false;

  PC++;                                                                         //byte 2

  if(isTagEnable(PC))
    noState = getNoValue();
  else{
    noState = true;
    PC++;                                                                       //byte 3
  }


  if(noState){
      int32_t tempSourceA = 0;
      int32_t tempSourceB = 0;

      PC++;                                                                     //byte 4
      tempSourceA = (int32_t)getSource();
      PC++;                                                                     //byte 6

      if(isTagEnable(PC+3)){                                                    //(byte 9)
        PC+=3;                                                                  //byte 9
        tempSourceB = (int32_t)getSource();
        PC++;                                                                   //byte 11
      }
      else{
        tempSourceB = (int32_t)getJoinValue();
        PC+=5;                                                                  //byte 11
      }

      tempSourceA = tempSourceA ^ tempSourceB;

      setDest(tempSourceA);
  }
  else
    PC+=9;
}//END xorOp()

void addOp(){                                                                   //Add
  bool noState = false;

  PC++;                                                                         //byte 2

  if(isTagEnable(PC))
    noState = getNoValue();
  else{
    noState = true;
    PC++;                                                                       //byte 3
  }


  if(noState){
      float tempSourceA = 0;
      float tempSourceB = 0;

      PC++;                                                                     //byte 4
      tempSourceA = getSource();
      PC++;                                                                     //byte 6

      if(isTagEnable(PC+3)){                                                    //(byte 9)
        PC+=3;                                                                  //byte 9
        tempSourceB = getSource();
        PC++;                                                                   //byte 11
      }
      else{
        tempSourceB = (float)getJoinValue();
        PC+=5;                                                                  //byte 11
      }

      tempSourceA += tempSourceB;

      setDest(tempSourceA);
      //Serial.println("ADD");
  }
  else
    PC+=9;
}

void subOp(){                                                                   //Sub
  bool noState = false;

  PC++;                                                                         //byte 2

  if(isTagEnable(PC))
    noState = getNoValue();
  else{
    noState = true;
    PC++;                                                                       //byte 3
  }


  if(noState){
      float tempSourceA = 0;
      float tempSourceB = 0;

      PC++;                                                                     //byte 4
      tempSourceA = getSource();
      PC++;                                                                     //byte 6

      if(isTagEnable(PC+3)){                                                    //(byte 9)
        PC+=3;                                                                  //byte 9
        tempSourceB = getSource();
        PC++;                                                                   //byte 11
      }
      else{
        tempSourceB = (float)getJoinValue();
        PC+=5;                                                                  //byte 11
      }

      tempSourceA -= tempSourceB;

      setDest(tempSourceA);
  }
  else
    PC+=9;
}

void mulOp(){                                                                   //Mul
  bool noState = false;

  PC++;                                                                         //byte 2

  if(isTagEnable(PC))
    noState = getNoValue();
  else{
    noState = true;
    PC++;                                                                       //byte 3
  }

  if(noState){
      float tempSourceA = 0;
      float tempSourceB = 0;

      PC++;                                                                     //byte 4
      tempSourceA = getSource();
      PC++;                                                                     //byte 6

      if(isTagEnable(PC+3)){                                                    //(byte 9)

        PC+=3;                                                                  //byte 9
        tempSourceB = getSource();

        PC++;                                                                   //byte 11
      }
      else{
        tempSourceB = (float)getJoinValue();

        PC+=5;                                                                  //byte 11
      }


      tempSourceA *= tempSourceB;

      setDest(tempSourceA);
      //Serial.println("MUL");
  }
  else
    PC+=9;
}

void divOp(){                                                                   //Div
  bool noState = false;

  PC++;                                                                         //byte 2

  if(isTagEnable(PC))
    noState = getNoValue();
  else{
    noState = true;
    PC++;                                                                       //byte 3
  }


  if(noState){
      float tempSourceA = 0;
      float tempSourceB = 0;

      PC++;                                                                     //byte 4
      tempSourceA = getSource();
      PC++;                                                                     //byte 6

      if(isTagEnable(PC+3)){                                                    //(byte 9)
        PC+=3;                                                                  //byte 9
        tempSourceB = getSource();
        PC++;                                                                   //byte 11
      }
      else{
        tempSourceB = (float)getJoinValue();
        PC+=5;                                                                  //byte 11
      }

      if(tempSourceB != 0)
        tempSourceA /= tempSourceB;

      setDest(tempSourceA);
  }
  else
    PC+=9;
}

void xpy(){                                                                     //x pow y
  bool noState = false;

  PC++;                                                                         //byte 2

  if(isTagEnable(PC))
    noState = getNoValue();
  else{
    noState = true;
    PC++;                                                                       //byte 3
  }

  if(noState){
      float tempSourceA = 0;
      float tempSourceB = 0;

      PC++;                                                                     //byte 4
      tempSourceA = getSource();
      PC++;                                                                     //byte 6

      if(isTagEnable(PC+3)){                                                    //(byte 9)
        PC+=3;                                                                  //byte 9
        tempSourceB = getSource();
        PC++;                                                                   //byte 11
      }
      else{
        tempSourceB = (float)getJoinValue();
        PC+=5;                                                                  //byte 11
      }

      tempSourceA = pow(tempSourceA, tempSourceB);

      setDest(tempSourceA);
  }
  else
    PC+=9;
}

void absMath(){                                                                 //Abs
  bool noState = false;

  PC++;                                                                         //byte 2

  if(isTagEnable(PC))
    noState = getNoValue();                                                     //(byte 3)
  else{
    noState = true;
    PC++;                                                                       //byte 3
  }

  if(noState){
      float tempSource = 0;

      PC++;                                                                     //byte 4
      tempSource = getSource();
      PC++;                                                                     //byte 6

      tempSource = abs(tempSource);

      setDest(tempSource);
  }
  else
    PC+=4;                                                                      //byte 7
}

void sqrMath(){
  bool noState = false;

  PC++;                                                                         //byte 2

  if(isTagEnable(PC))
    noState = getNoValue();                                                     //(byte 3)
  else{
    noState = true;
    PC++;                                                                       //byte 3
  }

  if(noState){
      float tempSource = 0;

      PC++;                                                                     //byte 4
      tempSource = getSource();
      PC++;                                                                     //byte 6

      tempSource = sqrt(tempSource);

      setDest(tempSource);
  }
  else
    PC+=4;                                                                      //byte 7
}

void asnMath(){
  bool noState = false;

  PC++;                                                                         //byte 2

  if(isTagEnable(PC))
    noState = getNoValue();                                                     //(byte 3)
  else{
    noState = true;
    PC++;                                                                       //byte 3
  }

  if(noState){
      float tempSource = 0;

      PC++;                                                                     //byte 4
      tempSource = getSource();
      PC++;                                                                     //byte 6

      tempSource = asin(tempSource);

      setDest(tempSource);
  }
  else
    PC+=4;                                                                      //byte 7
}

void acsMath(){
  bool noState = false;

  PC++;                                                                         //byte 2

  if(isTagEnable(PC))
    noState = getNoValue();                                                     //(byte 3)
  else{
    noState = true;
    PC++;                                                                       //byte 3
  }

  if(noState){
      float tempSource = 0;

      PC++;                                                                     //byte 4
      tempSource = getSource();
      PC++;                                                                     //byte 6

      tempSource = acos(tempSource);

      setDest(tempSource);
  }
  else
    PC+=4;                                                                      //byte 7
}

void atnMath(){
  bool noState = false;

  PC++;                                                                         //byte 2

  if(isTagEnable(PC))
    noState = getNoValue();                                                     //(byte 3)
  else{
    noState = true;
    PC++;                                                                       //byte 3
  }

  if(noState){
      float tempSource = 0;

      PC++;                                                                     //byte 4
      tempSource = getSource();
      PC++;                                                                     //byte 6

      tempSource = atan(tempSource);

      setDest(tempSource);
  }
  else
    PC+=4;                                                                      //byte 7
}

void cosMath(){
  bool noState = false;

  PC++;                                                                         //byte 2

  if(isTagEnable(PC))
    noState = getNoValue();                                                     //(byte 3)
  else{
    noState = true;
    PC++;                                                                       //byte 3
  }

  if(noState){
      float tempSource = 0;

      PC++;                                                                     //byte 4
      tempSource = getSource();
      PC++;                                                                     //byte 6

      tempSource = cos(tempSource);

      setDest(tempSource);
  }
  else
    PC+=4;                                                                      //byte 7
}

void lnMath(){
  bool noState = false;

  PC++;                                                                         //byte 2

  if(isTagEnable(PC))
    noState = getNoValue();                                                     //(byte 3)
  else{
    noState = true;
    PC++;                                                                       //byte 3
  }

  if(noState){
      float tempSource = 0;

      PC++;                                                                     //byte 4
      tempSource = getSource();
      PC++;                                                                     //byte 6

      tempSource = log(tempSource);

      setDest(tempSource);
  }
  else
    PC+=4;                                                                      //byte 7
}

void logMath(){
  bool noState = false;

  PC++;                                                                         //byte 2

  if(isTagEnable(PC))
    noState = getNoValue();                                                     //(byte 3)
  else{
    noState = true;
    PC++;                                                                       //byte 3
  }

  if(noState){
      float tempSource = 0;

      PC++;                                                                     //byte 4
      tempSource = getSource();
      PC++;                                                                     //byte 6

      tempSource = log10(tempSource);

      setDest(tempSource);
  }
  else
    PC+=4;                                                                      //byte 7
}

void sinMath(){
  bool noState = false;

  PC++;                                                                         //byte 2

  if(isTagEnable(PC))
    noState = getNoValue();                                                     //(byte 3)
  else{
    noState = true;
    PC++;                                                                       //byte 3
  }

  if(noState){
      float tempSource = 0;

      PC++;                                                                     //byte 4
      tempSource = getSource();
      PC++;                                                                     //byte 6

      tempSource = sin(tempSource);

      setDest(tempSource);
  }
  else
    PC+=4;                                                                      //byte 7
}

void tanMath(){
  bool noState = false;

  PC++;                                                                         //byte 2

  if(isTagEnable(PC))
    noState = getNoValue();                                                     //(byte 3)
  else{
    noState = true;
    PC++;                                                                       //byte 3
  }

  if(noState){
      float tempSource = 0;

      PC++;                                                                     //byte 4
      tempSource = getSource();
      PC++;                                                                     //byte 6

      tempSource = tan(tempSource);
      //Serial.print("tan: ");
      //Serial.println(tempSource);
      //Serial.println("");
      setDest(tempSource);
  }
  else
    PC+=4;                                                                      //byte 7
}

void bsl(){                                                                     //Bit Shift Left
  bool noState = false;

  PC++;                                                                         //byte 2

  if(isTagEnable(PC)){
    bool tempPre = getUpEdgeActive();
    bool noStateShift = getNoValue();

    if(tempPre)
      noState = true;

    if(!noStateShift){
        if(isTagEnable(PC+1))
          setDnValue(false, PC+1);
    }

  }
  else{
    noState = true;
    PC++;                                                                       //byte 3
  }


  if(noState){
    int dnIndex = 0;
    bool dnEnabled = false;
    int ulIndex = 0;
    bool ulEnabled = false;
    int32_t tempInt = 0;
    int sourceIndex = 0;
    bool bitSource = false;
    int32_t tempVarLength = 0;
    bool ulBit = false;

    PC++;                                                                       //byte 4
    if(isTagEnable(PC)){                                                        //if dn is enabled
      dnEnabled = true;
      dnIndex = PC;
    }

    PC+=2;                                                                      //byte 6

    if(isTagEnable(PC)){
      ulEnabled = true;
      ulIndex = PC;
    }

    PC+=2;                                                                      //byte 8

    bool typeVar = isTagEnable(PC);
    PC++;                                                                       //byte 9
    sourceIndex = PC;

    if(typeVar){                                                                //int-dint
      int32_t tempUl = 0;

      tempInt = (int32_t)getSource();
      PC++;                                                                     //byte 11

      if(isTagEnable(PC+1)){
        PC++;                                                                   //byte 12
        bitSource = getNoValue();
        PC++;                                                                   //byte 14
      }
      else{
        if(1==(int8_t)dueFlashStorage.read(PC))
          bitSource = true;
        PC+=3;                                                                  //byte 14
      }

      if(isTagEnable(PC+3)){
        PC+=3;                                                                  //byte 17
        tempVarLength = getSource();
      }
      else{
        tempVarLength = (int32_t)getJoinValue();
        PC+=4;                                                                  //byte 18
      }

      //Logic
      if(ulEnabled){
        tempUl = (tempInt & ((int32_t)1 << (tempVarLength - 1)));

        tempUl = (int32_t)(tempUl >> (tempVarLength - 1));
        if(tempUl == 1)
          ulBit = true;
        else
          ulBit = false;
      }

      int32_t maskUl = 1;
      for(int j = 0; j < (tempVarLength-1); j++){
        maskUl = (maskUl << 1);
        maskUl = (maskUl | 1);
      }

      tempUl = tempInt;

      tempUl = tempUl & maskUl;
      tempUl = tempUl << 1;
      tempUl = tempUl & maskUl;

      if(bitSource)
        tempUl = tempUl | 1;

      int32_t mask2 = ~maskUl;

      tempInt = (tempInt & mask2);

      tempInt = tempInt | tempUl;

      updateVariable(sourceIndex, tempInt);

    }//END int-dint

    else{                                                                       //output
      int outIndex = getOutIndex();
      PC++;                                                                     //byte 11

      if(isTagEnable(PC+1)){
        PC++;                                                                   //byte 12
        bitSource = getNoValue();
        PC++;                                                                   //byte 14
      }
      else{
        if(1==(int8_t)dueFlashStorage.read(PC))
          bitSource = true;
        PC+=3;                                                                  //byte 14
      }

      if(isTagEnable(PC+3)){
        PC+=3;                                                                  //byte 17
        tempVarLength = getSource();
      }
      else{
        tempVarLength = (int32_t)getJoinValue();
        PC+=4;                                                                  //byte 18
      }

      //Logic
      int tempCalc = 16 - outIndex;
      bool tempPostBit = false;
      bool newBit = bitSource;

      if(tempCalc >= 2){
        if(tempVarLength > tempCalc){

          if(ulEnabled)
            ulBit = oState[15];

          for(int j = outIndex; j < 16; j++){
            tempPostBit = oState[j];
            oState[j] = newBit;
            newBit = tempPostBit;
          }
        }

        else{
          if(ulEnabled)
            ulBit = oState[(outIndex + (tempVarLength-1))];

          for(int j = outIndex; j < (tempVarLength + outIndex); j++){
            tempPostBit = oState[j];
            oState[j] = newBit;
            newBit = tempPostBit;
          }
        }
      }//END if(tempCalc >= 2)
      else
        Serial.println("BSL Range Error");

    }//END output

    if(ulEnabled)
      setDnValue(ulBit, ulIndex);
    if(dnEnabled)
      setDnValue(true, dnIndex);

  }//END if(noState)
  else
    PC+=15;
}

void bsr(){                                                                     //Bit Shift Right
  bool noState = false;

  PC++;                                                                         //byte 2

  if(isTagEnable(PC)){
    bool tempPre = getUpEdgeActive();
    bool noStateShift = getNoValue();

    if(tempPre)
      noState = true;

    if(!noStateShift){
        if(isTagEnable(PC+1))
          setDnValue(false, PC+1);
    }

  }
  else{
    noState = true;
    PC++;                                                                       //byte 3
  }


  if(noState){
    int dnIndex = 0;
    bool dnEnabled = false;
    int ulIndex = 0;
    bool ulEnabled = false;
    int32_t tempInt = 0;
    int sourceIndex = 0;
    bool bitSource = false;
    int32_t tempVarLength = 0;
    bool ulBit = false;

    PC++;                                                                       //byte 4
    if(isTagEnable(PC)){                                                        //if dn is enabled
      dnEnabled = true;
      dnIndex = PC;
    }

    PC+=2;                                                                      //byte 6

    if(isTagEnable(PC)){
      ulEnabled = true;
      ulIndex = PC;
    }

    PC+=2;                                                                      //byte 8

    bool typeVar = isTagEnable(PC);
    PC++;                                                                       //byte 9
    sourceIndex = PC;

    if(typeVar){                                                                //int-dint
      int32_t tempUl = 0;

      tempInt = (int32_t)getSource();
      PC++;                                                                     //byte 11

      if(isTagEnable(PC+1)){
        PC++;                                                                   //byte 12
        bitSource = getNoValue();
        PC++;                                                                   //byte 14
      }
      else{
        if(1==(int8_t)dueFlashStorage.read(PC))
          bitSource = true;
        PC+=3;                                                                  //byte 14
      }

      if(isTagEnable(PC+3)){
        PC+=3;                                                                  //byte 17
        tempVarLength = getSource();
      }
      else{
        tempVarLength = (int32_t)getJoinValue();
        PC+=4;                                                                  //byte 18
      }



      //Logic
      if(ulEnabled){
        tempUl = (tempInt & 1);

        if(tempUl == 1)
          ulBit = true;
        else
          ulBit = false;
      }

      int32_t maskUl = 1;
      for(int j = 0; j < (tempVarLength-1); j++){
        maskUl = (maskUl << 1);
        maskUl = (maskUl | 1);
      }

      tempUl = tempInt;

      tempUl = tempUl & maskUl;
      tempUl = tempUl >> 1;
      tempUl = tempUl & maskUl;


      if(bitSource)
        tempUl = tempUl | ((int32_t)1 << (tempVarLength - 1));


      int32_t mask2 = ~maskUl;

      tempInt = (tempInt & mask2);

      tempInt = tempInt | tempUl;

      updateVariable(sourceIndex, tempInt);


    }//END int-dint

    else{                                                                       //output
      int outIndex = getOutIndex();
      PC++;                                                                     //byte 11

      if(isTagEnable(PC+1)){
        PC++;                                                                   //byte 12
        bitSource = getNoValue();
        PC++;                                                                   //byte 14
      }
      else{
        if(1==(int8_t)dueFlashStorage.read(PC))
          bitSource = true;
        PC+=3;                                                                  //byte 14
      }

      if(isTagEnable(PC+3)){
        PC+=3;                                                                  //byte 17
        tempVarLength = getSource();
      }
      else{
        tempVarLength = (int32_t)getJoinValue();
        PC+=4;                                                                  //byte 18
      }

      //Logic
      int tempCalc = 16 - outIndex;
      bool tempPostBit = false;
      bool newBit = bitSource;

      if(tempCalc >= 2){
        if(tempVarLength > tempCalc){

          if(ulEnabled)
            ulBit = oState[outIndex];

          for(int j = 16; j > outIndex; j--){
            tempPostBit = oState[j-1];
            oState[j-1] = newBit;
            newBit = tempPostBit;
          }
        }

        else{
          if(ulEnabled)
            ulBit = oState[outIndex];

          for(int j = (tempVarLength + outIndex); j > outIndex; j--){
            tempPostBit = oState[j-1];
            oState[j-1] = newBit;
            newBit = tempPostBit;
          }
        }
      }//END if(tempCalc >= 2)
      else
        Serial.println("BSL Range Error");

    }//END output

    if(ulEnabled)
      setDnValue(ulBit, ulIndex);
    if(dnEnabled)
      setDnValue(true, dnIndex);

  }//END if(noState)
  else
    PC+=15;
}

void rol(){                                                                     //Rotate left
  bool noState = false;

  PC++;                                                                         //byte 2

  if(isTagEnable(PC)){
    bool tempPre = getUpEdgeActive();
    bool noStateShift = getNoValue();

    if(tempPre)
      noState = true;

    if(!noStateShift){
        if(isTagEnable(PC+1))
          setDnValue(false, PC+1);
    }

  }
  else{
    noState = true;
    PC++;                                                                       //byte 3
  }


  if(noState){
    int dnIndex = 0;
    bool dnEnabled = false;
    int ulIndex = 0;
    bool ulEnabled = false;
    int32_t tempInt = 0;
    int sourceIndex = 0;
    int32_t rotate = 0;
    int32_t tempVarLength = 0;
    bool ulBit = false;

    PC++;                                                                       //byte 4
    if(isTagEnable(PC)){                                                        //if dn is enabled
      dnEnabled = true;
      dnIndex = PC;
    }

    PC+=2;                                                                      //byte 6

    if(isTagEnable(PC)){
      ulEnabled = true;
      ulIndex = PC;
    }

    PC+=2;                                                                      //byte 8

    bool typeVar = isTagEnable(PC);
    PC++;                                                                       //byte 9
    sourceIndex = PC;

    if(typeVar){                                                                //int-dint

      int32_t tempUl = 0;
      int32_t tempU2 = 0;

      tempInt = (int32_t)getSource();
      PC++;                                                                     //byte 11

      if(isTagEnable(PC+3)){
        PC+=3;                                                                  //byte 14
        rotate = (int32_t)getSource();
        PC++;                                                                   //byte 16
      }
      else{
        rotate = (int32_t)getJoinValue();
        PC+=5;                                                                  //byte 16
      }

      if(isTagEnable(PC+3)){
        PC+=3;                                                                  //byte 19
        tempVarLength = getSource();
      }
      else{
        tempVarLength = (int32_t)getJoinValue();
        PC+=4;                                                                  //byte 20
      }

      //Logic

      int32_t maskUl = 1;
      for(int j = 0; j < (tempVarLength-1); j++){
        maskUl = (maskUl << 1);
        maskUl = (maskUl | 1);
      }

      tempUl = tempInt;

      tempUl = tempUl & maskUl;

      int32_t tempPreBit = 0;

      for(int k=0; k<rotate; k++){
        tempU2 = (tempUl & ((int32_t)1 << (tempVarLength - 1)));
        tempPreBit = (tempU2 >> (tempVarLength - 1));

        tempUl = tempUl << 1;
        tempUl = tempUl & maskUl;
        tempUl = tempUl | tempPreBit;
      }

      if(ulEnabled){
        tempU2 = (tempUl & ((int32_t)1 << (tempVarLength - 1)));

        tempU2 = (tempU2 >> (tempVarLength - 1));
        if(tempU2 == 1)
          ulBit = true;
        else
          ulBit = false;
      }

      maskUl = ~maskUl;

      tempInt = (tempInt & maskUl);

      tempInt = tempInt | tempUl;

      updateVariable(sourceIndex, tempInt);


    }//END int-dint

    else{                                                                       //output
      int outIndex = getOutIndex();
      PC++;                                                                     //byte 11

      if(isTagEnable(PC+3)){
        PC+=3;                                                                  //byte 14
        rotate = (int32_t)getSource();
        PC++;                                                                   //byte 16
      }
      else{
        rotate = (int32_t)getJoinValue();
        PC+=5;                                                                  //byte 16
      }

      if(isTagEnable(PC+3)){
        PC+=3;                                                                  //byte 19
        tempVarLength = getSource();
      }
      else{
        tempVarLength = (int32_t)getJoinValue();
        PC+=4;                                                                  //byte 20
      }

      //Logic
      int tempCalc = 16 - outIndex;
      bool tempPostBit = false;

      if(tempCalc >= 2){
        if(tempVarLength > tempCalc){
          bool newBit = false;

          for(int k = 0; k < rotate; k++){
            for(int j = outIndex; j < 16; j++){
              tempPostBit = oState[j];
              oState[j] = newBit;
              newBit = tempPostBit;
            }
            oState[outIndex] = newBit;
          }

          if(ulEnabled)
            ulBit = oState[15];

        }

        else{

          bool newBit = false;

          for(int k = 0; k < rotate; k++){
            for(int j = outIndex; j < (tempVarLength + outIndex); j++){
              tempPostBit = oState[j];
              oState[j] = newBit;
              newBit = tempPostBit;
            }
            oState[outIndex] = newBit;
          }

          if(ulEnabled)
            ulBit = oState[(tempVarLength + (outIndex-1))];

        }
      }//END if(tempCalc >= 2)
      else
        Serial.println("ROL Range Error");

    }//END output

    if(ulEnabled)
      setDnValue(ulBit, ulIndex);
    if(dnEnabled)
      setDnValue(true, dnIndex);

  }//END if(noState)
  else
    PC+=17;
}

void ror(){                                                                     //Rotate Right
  bool noState = false;

  PC++;                                                                         //byte 2

  if(isTagEnable(PC)){
    bool tempPre = getUpEdgeActive();
    bool noStateShift = getNoValue();

    if(tempPre)
      noState = true;

    if(!noStateShift){
        if(isTagEnable(PC+1))
          setDnValue(false, PC+1);
    }

  }
  else{
    noState = true;
    PC++;                                                                       //byte 3
  }


  if(noState){
    int dnIndex = 0;
    bool dnEnabled = false;
    int ulIndex = 0;
    bool ulEnabled = false;
    int32_t tempInt = 0;
    int sourceIndex = 0;
    int32_t rotate = 0;
    int32_t tempVarLength = 0;
    bool ulBit = false;

    PC++;                                                                       //byte 4
    if(isTagEnable(PC)){                                                        //if dn is enabled
      dnEnabled = true;
      dnIndex = PC;
    }

    PC+=2;                                                                      //byte 6

    if(isTagEnable(PC)){
      ulEnabled = true;
      ulIndex = PC;
    }

    PC+=2;                                                                      //byte 8

    bool typeVar = isTagEnable(PC);
    PC++;                                                                       //byte 9
    sourceIndex = PC;

    if(typeVar){                                                                //int-dint

      int32_t tempUl = 0;
      int32_t tempU2 = 0;

      tempInt = (int32_t)getSource();
      PC++;                                                                     //byte 11

      if(isTagEnable(PC+3)){
        PC+=3;                                                                  //byte 14
        rotate = (int32_t)getSource();
        PC++;                                                                   //byte 16
      }
      else{
        rotate = (int32_t)getJoinValue();
        PC+=5;                                                                  //byte 16
      }

      if(isTagEnable(PC+3)){
        PC+=3;                                                                  //byte 19
        tempVarLength = getSource();
      }
      else{
        tempVarLength = (int32_t)getJoinValue();
        PC+=4;                                                                  //byte 20
      }

      //Logic

      int32_t maskUl = 1;
      for(int j = 0; j < (tempVarLength-1); j++){
        maskUl = (maskUl << 1);
        maskUl = (maskUl | 1);
      }

      tempUl = tempInt;

      tempUl = tempUl & maskUl;

      int32_t tempPreBit = 0;

      for(int k=0; k<rotate; k++){
        tempPreBit =  (tempUl & 1);
        tempPreBit =  (tempPreBit << (tempVarLength -1));

        tempUl = tempUl >> 1;
        tempUl = tempUl & maskUl;
        tempUl = tempUl | tempPreBit;
      }

      if(ulEnabled){
        tempU2 = (tempPreBit >> (tempVarLength - 1));
        if(tempU2 == 1)
          ulBit = true;
        else
          ulBit = false;
      }

      maskUl = ~maskUl;

      tempInt = (tempInt & maskUl);

      tempInt = tempInt | tempUl;

      updateVariable(sourceIndex, tempInt);


    }//END int-dint

    else{                                                                       //output
      int outIndex = getOutIndex();
      PC++;                                                                     //byte 11

      if(isTagEnable(PC+3)){
        PC+=3;                                                                  //byte 14
        rotate = (int32_t)getSource();
        PC++;                                                                   //byte 16
      }
      else{
        rotate = (int32_t)getJoinValue();
        PC+=5;                                                                  //byte 16
      }

      if(isTagEnable(PC+3)){
        PC+=3;                                                                  //byte 19
        tempVarLength = getSource();
      }
      else{
        tempVarLength = (int32_t)getJoinValue();
        PC+=4;                                                                  //byte 20
      }

      //Logic
      int tempCalc = 16 - outIndex;
      bool tempPostBit = false;

      if(tempCalc >= 2){
        if(tempVarLength > tempCalc){
          bool newBit = false;

          for(int k = 0; k < rotate; k++){
            for(int j = 16; j > outIndex; j--){
              tempPostBit = oState[j-1];
              oState[j-1] = newBit;
              newBit = tempPostBit;
            }
            oState[15] = newBit;
          }

          if(ulEnabled)
            ulBit = oState[15];

        }

        else{
          bool newBit = false;

          for(int k = 0; k < rotate; k++){
            for(int j = (tempVarLength + outIndex); j > outIndex; j--){
              tempPostBit = oState[j-1];
              oState[j-1] = newBit;
              newBit = tempPostBit;
            }
            oState[tempVarLength + (outIndex-1)] = newBit;
          }

          if(ulEnabled)
            ulBit = oState[(tempVarLength + (outIndex-1))];

        }
      }//END if(tempCalc >= 2)
      else
        Serial.println("ROR Range Error");

    }//END output

    if(ulEnabled)
      setDnValue(ulBit, ulIndex);
    if(dnEnabled)
      setDnValue(true, dnIndex);

  }//END if(noState)
  else
    PC+=17;
}

void scl(){                                                                     //Scale
  bool noState = false;

  PC++;                                                                         //byte 2

  if(isTagEnable(PC))
    noState = getNoValue();
  else{
    noState = true;
    PC++;                                                                       //byte 3
  }


  if(noState){
      int outputIndex = 0;
      float tempInput = 0;
      float tempRate = 0;
      float tempOffset = 0;


      PC++;                                                                     //byte 4
      outputIndex = PC;
      PC+=2;                                                                    //byte 6

      tempInput = getSource();
      PC++;                                                                     //byte 8

      if(isTagEnable(PC+3)){
        PC+=3;                                                                  //byte 11
        tempRate = getSource();
        PC++;                                                                   //byte 13
      }
      else{
        tempRate = (float)getJoinValue();
        PC+=5;                                                                  //byte 13
      }

      if(isTagEnable(PC+3)){
        PC+=3;                                                                  //byte 16
        tempOffset = getSource();
      }
      else{
        tempOffset = (float)getJoinValue();
        PC+=4;                                                                  //byte 17
      }

      float tempResult = tempRate / 10000L;
      tempResult *= tempInput;
      tempResult += tempOffset;


      setDest(outputIndex, tempResult);


  }
  else
    PC+=14;
}

void scp(){                                                                     //Scale with parameter
  bool noState = false;

  PC++;                                                                         //byte 2

  if(isTagEnable(PC))
    noState = getNoValue();
  else{
    noState = true;
    PC++;                                                                       //byte 3
  }


  if(noState){
      int outputIndex = 0;
      float tempInput = 0;
      float tempInputMin = 0;
      float tempInputMax = 0;
      float tempOutputMin = 0;
      float tempOutputMax = 0;

      PC++;                                                                     //byte 4
      outputIndex = PC;
      PC+=2;                                                                    //byte 6

      tempInput = getSource();
      PC++;                                                                     //byte 8

      if(isTagEnable(PC+3)){
        PC+=3;                                                                  //byte 11
        tempInputMin = getSource();
        PC++;                                                                   //byte 13
      }
      else{
        tempInputMin = (float)getJoinValue();
        PC+=5;                                                                  //byte 13
      }

      if(isTagEnable(PC+3)){
        PC+=3;                                                                  //byte 16
        tempInputMax = getSource();
        PC++;                                                                   //byte 18
      }
      else{
        tempInputMax = (float)getJoinValue();
        PC+=5;                                                                  //byte 18
      }


      if(isTagEnable(PC+3)){
        PC+=3;                                                                  //byte 21
        tempOutputMin = getSource();
        PC++;                                                                   //byte 23
      }
      else{
        tempOutputMin = (float)getJoinValue();
        PC+=5;                                                                  //byte 23
      }

      if(isTagEnable(PC+3)){
        PC+=3;                                                                  //byte 26
        tempOutputMax = getSource();
      }
      else{
        tempOutputMax = (float)getJoinValue();
        PC+=4;                                                                  //byte 27
      }

      float tempResult = 0;
      bool calculatePlease = true;

      if(tempInputMax == tempInputMin)
        calculatePlease = false;
      if(tempOutputMin == tempOutputMax)
        calculatePlease = false;
      if(tempInputMin > tempInputMax)
        calculatePlease = false;
      if(tempOutputMin > tempOutputMax)
        calculatePlease = false;

      if(calculatePlease){
        float tempRateCalc =((tempOutputMax - tempOutputMin) / (tempInputMax - tempInputMin));
        float tempOffsetCalc = tempOutputMin - (tempInputMin * tempRateCalc);

        tempResult = (tempInput * tempRateCalc) + tempOffsetCalc;
      }
      else
        tempResult = tempOutputMin;

      if(tempResult > tempOutputMax)
        tempResult = tempOutputMax;
      else if(tempResult < tempOutputMin)
        tempResult = tempOutputMin;

      setDest(outputIndex, tempResult);
      //Serial.println("SCP");


  }
  else
    PC+=24;
}

bool isTagEnable(int tempPC){
  bool tempBool = false;

  if(dueFlashStorage.read(tempPC) != 0)
    tempBool = true;

  return tempBool;
}

int16_t getJoinValue(){
  int16_t tempA = 0;
  int16_t tempB = 0;

  tempB = (int16_t)dueFlashStorage.read(PC) << 14;
  tempA = (int16_t)dueFlashStorage.read(PC+1) << 7;
  tempB = tempB | tempA | (int16_t)dueFlashStorage.read(PC+2);

  return tempB;
}

int getOutIndex(){
  int tempReturn = 0;

  uint8_t myDataByte = dueFlashStorage.read(PC);
  PC++;
  uint8_t n = dueFlashStorage.read(PC);
  n = n - 48;

  switch(myDataByte){
    case 'O':                                                                   //output
      tempReturn = n;
    break;
    case 'k':                                                                   //output + 10
      tempReturn = n+10;
    break;
  }

  return tempReturn;
}

void setDnValue(bool newState, int tempPC){

  uint8_t myDataByte = dueFlashStorage.read(tempPC);
  uint8_t n = dueFlashStorage.read(tempPC+1);
  n = n-48;

  switch(myDataByte){
    case 'O':                                                                   //output
      oState[n] = newState;
    break;
    case 'k':                                                                   //output + 10
    oState[n+10] = newState;
    break;
    case 'V':                                                                   //virtual
      vState[n] = newState;
    break;
    case 'v':                                                                   //virtual + 10
      vState[n+10] = newState;
    break;
    case 'U':                                                                   //virtual + 20
      vState[n+20] = newState;
    break;
    case 'u':                                                                   //virtual + 30
      vState[n+30] = newState;
    break;
    case 'A':                                                                   //virtual + 40
      vState[n+40] = newState;
    break;
    case 'a':                                                                   //virtual + 50
      vState[n+50] = newState;
    break;
    case 'c':                                                                   //virtual + 60
      vState[n+60] = newState;
    break;
  }

}

bool getUpEdgeActive(){
  bool tempReturn = false;

  uint8_t myDataByte = dueFlashStorage.read(PC);
  uint8_t n = dueFlashStorage.read(PC+1);
  n = n - 48;

  if((myDataByte) != 0){
    switch(myDataByte){
      case 'I':                                                                 //intput
        tempReturn = iUpEdge[n];
      break;
      case 'i':                                                                 //input + 10
        tempReturn = iUpEdge[n+10];
      break;
      case 'O':                                                                   //output
        tempReturn = oUpEdge[n];
      break;
      case 'k':                                                                   //output + 10
        tempReturn = oUpEdge[n+10];
      break;
      case 'V':                                                                   //virtual
        tempReturn = vUpEdge[n];
      break;
      case 'v':                                                                   //virtual + 10
        tempReturn = vUpEdge[n+10];
      break;
      case 'U':                                                                   //virtual + 20
        tempReturn = vUpEdge[n+20];
      break;
      case 'u':                                                                   //virtual + 30
        tempReturn = vUpEdge[n+30];
      break;
      case 'A':                                                                   //virtual + 40
        tempReturn = vUpEdge[n+40];
      break;
      case 'a':                                                                   //virtual + 50
        tempReturn = vUpEdge[n+50];
      break;
      case 'c':                                                                   //virtual + 60
        tempReturn = vUpEdge[n+60];
      break;

    }
  }
  else
    tempReturn = true;

  return tempReturn;
}

bool getNoValue(){
  bool tempReturn = false;

  uint8_t myDataByte = dueFlashStorage.read(PC);
  PC++;
  uint8_t n = dueFlashStorage.read(PC);
  n = n - 48;


  if((myDataByte) != 0){
    switch(myDataByte){
      case 'I':                                                                 //intput
        tempReturn = iState[n];
      break;
      case 'i':                                                                 //input + 10
        tempReturn = iState[n+10];
      break;
      case 'O':                                                                   //output
        tempReturn = oState[n];
      break;
      case 'k':                                                                   //output + 10
        tempReturn = oState[n+10];
      break;
      case 'V':                                                                   //virtual
        tempReturn = vState[n];
      break;
      case 'v':                                                                   //virtual + 10
        tempReturn = vState[n+10];
      break;
      case 'U':                                                                   //virtual + 20
        tempReturn = vState[n+20];
      break;
      case 'u':                                                                   //virtual + 30
        tempReturn = vState[n+30];
      break;
      case 'A':                                                                   //virtual + 40
        tempReturn = vState[n+40];
      break;
      case 'a':                                                                   //virtual + 50
        tempReturn = vState[n+50];
      break;
      case 'c':                                                                   //virtual + 60
        tempReturn = vState[n+60];
      break;

    }
  }
  else
    tempReturn = true;

  return tempReturn;
}

bool isSourceInt(){
  bool tempReturn = false;

  uint8_t myDataByte = dueFlashStorage.read(PC);

  switch (myDataByte) {
    case 'D':                                                                   //int
      tempReturn = true;
    break;
    case 'd':                                                                   //int + 10
      tempReturn = true;
    break;
    case 'M':                                                                   //int + 20
      tempReturn = true;
    break;
    case 'N':                                                                   //int + 30
      tempReturn = true;
    break;
    case 'y':                                                                   //int time                                    
      tempReturn = true;
    break;
  }

  return tempReturn;
}

float getSource(){
  float tempReturn = 0;

  uint8_t myDataByte = dueFlashStorage.read(PC);
  PC++;
  uint8_t n = dueFlashStorage.read(PC);
  n = n - 48;

  switch(myDataByte){
    case 'p':                                                                   //analog input
      tempReturn = (float)readAnalogPin(n);
    break;
    case 'Q':                                                                   //analog input + 10
      tempReturn = (float)readAnalogPin(n+10);
    break;
    case 'o':                                                                   //analog output
      tempReturn = (float)readAnalogOutput(n);
    break;
    case 'q':                                                                   //analog output + 10
      tempReturn = (float)readAnalogOutput(n+10);
    break;
    case 'D':                                                                   //int
      tempReturn = (float)intValues[n];
    break;
    case 'd':                                                                   //int + 10
      tempReturn = (float)intValues[n+10];
    break;
    case 'M':                                                                   //int + 20
      tempReturn = (float)intValues[n+20];
    break;
    case 'N':                                                                   //int + 30
      tempReturn = (float)intValues[n+30];
    break;
    case 'y':
      tempReturn = (float)timeData[n];                                          //int time                                        
    break;
    case 'L':                                                                   //long
      tempReturn = (float)dintValues[n];
    break;
    case 'H':                                                                   //long + 10
      tempReturn = (float)dintValues[n+10];
    break;
    case 'h':                                                                   //long + 20
      tempReturn = (float)dintValues[n+20];
    break;
    case 'J':                                                                   //long + 30
      tempReturn = (float)dintValues[n+30];
    break;
    case 'j':                                                                   //long + 40
      tempReturn = (float)dintValues[n+40];
    break;
    case 'K':                                                                   //long + 50
      tempReturn = (float)dintValues[n+50];
    break;
    case 'F':                                                                   //real
      tempReturn = realValues[n];
    break;
    case 'E':                                                                   //real + 10
      tempReturn = realValues[n+10];
    break;
    case 'g':                                                                   //real + 20
      tempReturn = realValues[n+20];
    break;
    case 'G':                                                                   //real + 30
      tempReturn = realValues[n+30];
    break;
    default:
      tempReturn = 0;
      break;
  }

  return tempReturn;
}

void updateVariable(int tempPC, int32_t valueDest){
  uint8_t myDataByte = dueFlashStorage.read(tempPC);
  uint8_t n = dueFlashStorage.read(tempPC+1);
  n = n-48;

  switch(myDataByte){
        case 'D':                                                                   //int
          intValues[n] = (int16_t)valueDest;
        break;
        case 'd':                                                                   //int + 10
          intValues[n+10] = (int16_t)valueDest;
        break;
        case 'M':                                                                   //int + 20
          intValues[n+20] = (int16_t)valueDest;
        break;
        case 'N':                                                                   //int + 30
          intValues[n+30] = (int16_t)valueDest;
        break;
        case 'L':                                                                   //long
          dintValues[n] = valueDest;
        break;
        case 'H':                                                                   //long + 10
          dintValues[n+10] = valueDest;
        break;
        case 'h':                                                                   //long + 20
          dintValues[n+20] = valueDest;
        break;
        case 'J':                                                                   //long + 30
          dintValues[n+30] = valueDest;
        break;
        case 'j':                                                                   //long + 40
          dintValues[n+40] = valueDest;
        break;
        case 'K':                                                                   //long + 50
          dintValues[n+50] = valueDest;
        break;

  }
}

void setDest(float valueDest){

  uint8_t myDataByte = dueFlashStorage.read(PC);
  PC++;
  uint8_t n = dueFlashStorage.read(PC);
  n = n-48;

  switch(myDataByte){
        case 'x':
            writeServoPin(n, (int)valueDest);
        break;
        case 'X':
            writeServoPin(n+10, (int)valueDest);
        break;
        case 'o':                                                                   //Analog output
          writeAnalogPin(n, (int)valueDest);
        break;
        case 'q':                                                                   //Analog output + 10
          writeAnalogPin(n+10, (int)valueDest);
        break;
        case 'D':                                                                   //int
          intValues[n] = (int16_t)valueDest;
        break;
        case 'd':                                                                   //int + 10
          intValues[n+10] = (int16_t)valueDest;
        break;
        case 'M':                                                                   //int + 20
          intValues[n+20] = (int16_t)valueDest;
        break;
        case 'N':                                                                   //int + 30
          intValues[n+30] = (int16_t)valueDest;
        break;
        case 'L':                                                                   //long
          dintValues[n] = (int32_t)valueDest;
        break;
        case 'H':                                                                   //long + 10
          dintValues[n+10] = (int32_t)valueDest;
        break;
        case 'h':                                                                   //long + 20
          dintValues[n+20] = (int32_t)valueDest;
        break;
        case 'J':                                                                   //long + 30
          dintValues[n+30] = (int32_t)valueDest;
        break;
        case 'j':                                                                   //long + 40
          dintValues[n+40] = (int32_t)valueDest;
        break;
        case 'K':                                                                   //long + 50
          dintValues[n+50] = (int32_t)valueDest;
        break;
        case 'F':                                                                   //real
          realValues[n] = valueDest;
        break;
        case 'E':                                                                   //real + 10
          realValues[n+10] = valueDest;
        break;
        case 'g':                                                                   //real + 20
          realValues[n+20] = valueDest;
        break;
        case 'G':                                                                   //real + 30
          realValues[n+30] = valueDest;
        break;

  }

}

void setDest(int tempPC, float valueDest){

  uint8_t myDataByte = dueFlashStorage.read(tempPC);
  uint8_t n = dueFlashStorage.read(tempPC+1);
  n = n-48;

  switch(myDataByte){
        case 'x':
            writeServoPin(n, (int)valueDest);
        break;
        case 'X':
            writeServoPin(n+10, (int)valueDest);
        break;
        case 'o':                                                                   //Analog output
          writeAnalogPin(n, (int)valueDest);
        break;
        case 'q':                                                                   //Analog output + 10
          writeAnalogPin(n+10, (int)valueDest);
        break;
        case 'D':                                                                   //int
          intValues[n] = (int16_t)valueDest;
        break;
        case 'd':                                                                   //int + 10
          intValues[n+10] = (int16_t)valueDest;
        break;
        case 'M':                                                                   //int + 20
          intValues[n+20] = (int16_t)valueDest;
        break;
        case 'N':                                                                   //int + 30
          intValues[n+30] = (int16_t)valueDest;
        break;
        case 'L':                                                                   //long
          dintValues[n] = (int32_t)valueDest;
        break;
        case 'H':                                                                   //long + 10
          dintValues[n+10] = (int32_t)valueDest;
        break;
        case 'h':                                                                   //long + 20
          dintValues[n+20] = (int32_t)valueDest;
        break;
        case 'J':                                                                   //long + 30
          dintValues[n+30] = (int32_t)valueDest;
        break;
        case 'j':                                                                   //long + 40
          dintValues[n+40] = (int32_t)valueDest;
        break;
        case 'K':                                                                   //long + 50
          dintValues[n+50] = (int32_t)valueDest;
        break;
        case 'F':                                                                   //real
          realValues[n] = valueDest;
        break;
        case 'E':                                                                   //real + 10
          realValues[n+10] = valueDest;
        break;
        case 'g':                                                                   //real + 20
          realValues[n+20] = valueDest;
        break;
        case 'G':                                                                   //real + 30
          realValues[n+30] = valueDest;
        break;

  }

}


void fragmentData(uint32_t tempInt32, uint8_t basura){
  ld = (tempInt32 & (BYTE6MASK0));

  tempValue32 = (tempInt32 & (BYTE6MASK1));
  md = (tempValue32 >> 6);

  tempValue32 = (tempInt32 & (BYTE6MASK2));
  hd = (tempValue32 >> 12);

  tempValue32 = (tempInt32 & (BYTE6MASK3));
  ld32 = (tempValue32 >>18);

  tempValue32 = (tempInt32 & (BYTE6MASK4));
  md32 = (tempValue32 >> 24);

  tempValue32 = (tempInt32 & (BYTE6MASK5));
  hd32 = (tempValue32 >> 30);

}

void fragmentData(uint32_t tempInt16){
  ld = (tempInt16 & (BYTE6MASK0));

  tempValue = (tempInt16 & (MDMASK));
  md = (tempValue >> 6);

  tempValue = (tempInt16 & (HDMASK));
  hd = (tempValue >> 12);
}


void fragmentData(uint32_t tempInt16, boolean bol, boolean bol2){
  ld = (tempInt16 & (BYTE6MASK0));

  tempValue = (tempInt16 & (MDMASK));
  md = (tempValue >> 6);
}

void checkSpace(){
  PC=0; //put the PC in the start of the memory
  int dataByte= dueFlashStorage.read(PC);
  Serial.println("");
  Serial.write(dataByte);
  while(dataByte!='z'){ //z is the value that indicate that the program has ended
    PC++;
    dataByte= dueFlashStorage.read(PC);
    if((dataByte == 'R')|| (dataByte == '$'))
      Serial.println("");
    Serial.write(dataByte);
  }

  dataByte = PC+1;
  int tempPC = (100 * (dataByte))/REG_TIMER;


  Serial.println("");
  Serial.print("Program bytes: ");
  Serial.print(PC+1);
  Serial.print(" (");
  Serial.print(tempPC);
  Serial.println("%)");


/*
  Serial.println("");
  Serial.println("Int:");
  Serial.write(dueFlashStorage.read(4057));
  Serial.println("Dint:");
  Serial.write(dueFlashStorage.read(4058));
  Serial.println("Real:");
  Serial.write(dueFlashStorage.read(4059));
  Serial.println("");
  */

  Serial.print("Free: ");
  tempPC = 100 - tempPC;
  Serial.print(tempPC);
  Serial.println("%");
  Serial.println(" ");
}



int setGMT(int timeOffset){
  int secondsOffset = 0;
  boolean negativeValue = false;

  if(timeOffset < 0){
    negativeValue = true;
    timeOffset *= (-1);
  }

  switch(timeOffset){
    case 100:         //GMT+1:00
      secondsOffset = 3600;
      break;
    case 200:         //GMT+2:00
      secondsOffset = 7200;
      break;
    case 230:         //GMT+2:30
      secondsOffset = 9000;
      break;
    case 300:         //GMT+3:00
      secondsOffset = 10800;
      break;
    case 400:         //GMT+4:00
      secondsOffset = 14400;
      break;
    case 430:         //GMT+4:30
      secondsOffset = 16200;
      break;   
    case 500:         //GMT+5:00
      secondsOffset = 18000;
      break;
    case 530:         //GMT+5:30
      secondsOffset = 19800;
      break;  
    case 545:         //GMT+5:45
      secondsOffset = 20700;
      break;
    case 600:         //GMT+6:00
      secondsOffset = 21600;
      break;
    case 630:         //GMT+6:30
      secondsOffset = 23400;
      break;   
    case 700:         //GMT+7:00
      secondsOffset = 25200;
      break;
    case 800:         //GMT+8:00
      secondsOffset = 28800;
      break;
    case 830:         //GMT+8:30
      secondsOffset = 30600;
      break;
    case 845:         //GMT+8:45
      secondsOffset = 31500;
      break;  
    case 900:         //GMT+9:00
      secondsOffset = 32400;
      break;
    case 930:         //GMT+9:30
      secondsOffset = 34200;
      break;
    case 1000:        //GMT+10:00
      secondsOffset = 36000;
      break;
    case 1030:        //GMT+10:30
      secondsOffset = 37800;
      break;   
    case 1100:        //GMT+11:00
      secondsOffset = 39600;
      break;
    case 1200:        //GMT+12:00    
      secondsOffset = 43200;     
      break;
    case 1245:        //GMT+12:45
      secondsOffset = 45900;
      break;
    case 1300:        //GMT+13:00
      secondsOffset = 46800;
      break;
    case 1400:        //GMT+14:00
      secondsOffset = 50400;
      break;      
  }

  if(negativeValue)
    secondsOffset *= (-1);
   

  return secondsOffset;
}


void getTime(){
  if(enabledRTC){
    setTime(RTC.get());
  }
}



void updateTimeRegisters(){
  timeData[0] = year();
  timeData[1] = month();
  timeData[2] = weekday();
  timeData[3] = day();
  timeData[4] = hour();
  timeData[5] = minute();
  timeData[6] = second();
}

void updateTimeFromRTC(){
  if(enabledRTC){
    unsigned long tempRTCTime = millis();
    tempRTCTime -= currentGetRTCUpdateTime;
    int currentTime = (int32_t)(tempRTCTime * (0.001f));

    if(currentTime < 0){
      currentGetRTCUpdateTime = millis();
      currentTime *= (-1);
    }

    if(currentTime >= targetGetRTCUpdateTime){
      
      currentGetRTCUpdateTime = millis();

      if(RTC.get() != 0){
        setTime(RTC.get());
      }
    }  
  }
}

void updateAllTime(){
  setTime(timeData[4], timeData[5], timeData[6], timeData[3], timeData[1]+1, timeData[0]);
  
  if(enabledRTC){
    RTC.set(now());
    currentGetRTCUpdateTime = millis();
  }
}


void swapLState(){
  bool tempSate = digitalRead(13);
  if(tempSate)
    digitalWrite(13, LOW);
  else
    digitalWrite(13, HIGH);  
}



uint8_t joinInt8(char tempMd, char tempLd){
  uint8_t temp = 0;

  uint8_t tempLd1 = (int)tempLd;
  uint8_t tempMd1 = (int)tempMd;

  tempMd1 = (tempMd1 << 6);

  temp = (tempLd1 | tempMd1);
  temp = temp & GMASK8;

  return temp;
}


uint16_t joinInt16(char tempHd, char tempMd, char tempLd){
  uint16_t temp = 0;

  uint16_t tempLd1 = (int)tempLd;
  uint16_t tempMd1 = (int)tempMd;
  uint16_t tempHd1 = (int)tempHd;

  tempMd1 = (tempMd1 << 6);
  tempHd1 = (tempHd1 << 12);

  temp = (tempLd1 | tempMd1 | tempHd1);
  temp = temp & GMASK16;

  return temp;
}

uint32_t joinInt32(char temp6, char temp5, char temp4, char temp3, char temp2, char temp1){
  uint32_t tempA = 0;
  uint32_t tempB = 0;

  tempB = (((int32_t)temp6) << 30);
  tempA = (((int32_t)temp5) << 24);
  tempB = tempB | tempA;
  tempA = (((int32_t)temp4) << 18);
  tempB = tempB | tempA;
  tempA = (((int32_t)temp3) << 12);
  tempB = tempB | tempA;
  tempA = (((int32_t)temp2) << 6);
  tempB = tempB | tempA | ((int32_t)temp1);

  return tempB;
}

float joinFloat(uint32_t tempInt32){
  float temp = 0;
  floatValue tempFloat;

  tempFloat.bytes[0] = (uint8_t)(tempInt32 & BYTEMASK0);
  tempFloat.bytes[1] = (uint8_t)((tempInt32 & BYTEMASK1) >> 8);
  tempFloat.bytes[2] = (uint8_t)((tempInt32 & BYTEMASK2) >> 16);
  tempFloat.bytes[3] = (uint8_t)((tempInt32 & BYTEMASK3) >> 24);

  temp = tempFloat.val;

  return temp;
}








